"""Instructions common to all ``PushTypes`` and ``PushStacks``."""
from typing import Union, Callable

from pyshgp.push.instruction import (
    SimpleInstruction,
    TakesStateInstruction,
    StateToStateInstruction,
    ProducesManyOfTypeInstruction
)
from pyshgp.push.state import PushState
from pyshgp.utils import Token


DUP_LIMIT = 500


def _dup_times(times, item):
    times = min(times, DUP_LIMIT)
    return [item] * times


# Disabled due to performance issues.
# def _dup_top_n_factory(type_name: str) -> Callable:
#     def f(state: PushState) -> PushState:
#         if state["int"].is_empty() or state[type_name].is_empty():
#             return Token.revert
#         ndx = max(0, min(state["int"].pop(), len(state[type_name]) - 1))
#         for item in state[type_name][:ndx]:
#             state[type_name].push(item)
#         return state
#     return f


def _flusher(type_name: str) -> Callable:
    def f(state: PushState) -> PushState:
        state[type_name].flush()
        return state
    return f


def _stack_depther(type_name: str) -> Callable:
    def f(state: PushState) -> int:
        return [len(state[type_name])]
    return f


def _yanker(type_name: str) -> Callable:
    def f(state: PushState) -> Union[PushState, Token]:
        if state["int"].is_empty() or state[type_name].is_empty():
            return Token.revert
        if type_name == "int" and len(state[type_name]) < 2:
            return Token.revert
        raw_ndx = state["int"].pop()
        ndx = max(0, min(raw_ndx, len(state[type_name]) - 1))
        item = state[type_name].pop(ndx)
        state[type_name].push(item)
        return state
    return f


def _yank_duper(type_name: str) -> Callable:
    def f(state: PushState) -> Union[PushState, Token]:
        if state["int"].is_empty() or state[type_name].is_empty():
            return Token.revert
        if type_name == "int" and len(state[type_name]) < 2:
            return Token.revert
        raw_ndx = state["int"].pop()
        ndx = max(0, min(raw_ndx, len(state[type_name]) - 1))
        item = state[type_name].nth(ndx)
        state[type_name].push(item)
        return state
    return f


def _shover(type_name: str) -> Callable:
    def f(state: PushState) -> Union[PushState, Token]:
        if state["int"].is_empty() or state[type_name].is_empty():
            return Token.revert
        if type_name == "int" and len(state[type_name]) < 2:
            return Token.revert
        raw_ndx = state["int"].pop()
        ndx = max(0, min(raw_ndx, len(state[type_name]) - 1))
        item = state[type_name].pop()
        state[type_name].insert(ndx, item)
        return state
    return f


def _shove_duper(type_name: str) -> Callable:
    def f(state: PushState) -> Union[PushState, Token]:
        if state["int"].is_empty() or state[type_name].is_empty():
            return Token.revert
        if type_name == "int" and len(state[type_name]) < 2:
            return Token.revert
        raw_ndx = state["int"].pop()
        ndx = max(0, min(raw_ndx, len(state[type_name])))
        item = state[type_name].top()
        state[type_name].insert(ndx, item)
        return state
    return f


def _is_emptyer(type_name: str) -> Callable:
    def f(state: PushState) -> bool:
        return [state[type_name].is_empty()]
    return f


def instructions():
    """Return all core numeric instructions."""
    i = []

    for push_type in ["bool", "int", "float", "str", "char", "code", "exec"]:
        i.append(SimpleInstruction(
            "{t}_pop".format(t=push_type),
            lambda x: [],
            input_types=[push_type],
            output_types=[],
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Pops the top {t}.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_dup".format(t=push_type),
            lambda x: [x, x],
            input_types=[push_type],
            output_types=[push_type, push_type],
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Duplicates the top {t}.".format(t=push_type)
        ))

        i.append(ProducesManyOfTypeInstruction(
            "{t}_dup_times".format(t=push_type),
            _dup_times,
            input_types=["int", push_type],
            output_type=push_type,
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Duplicates the top {t} `n` times where `n` is from the int stack.".format(t=push_type)
        ))

        # Disabled due to performance issues.
        # i.append(StateToStateInstruction(
        #     "{t}_dup_top_n".format(t=push_type),
        #     _dup_top_n_factory(push_type),
        #     types_used=[push_type, "int"],
        #     code_blocks=0,
        #     docstring="Duplicates the top n items on the {t} stack.".format(t=push_type)
        # ))

        i.append(SimpleInstruction(
            "{t}_swap".format(t=push_type),
            lambda a, b: [a, b],
            input_types=[push_type, push_type],
            output_types=[push_type, push_type],
            code_blocks=(2 if push_type == "exec" else 0),
            docstring="Swaps the top two {t}s.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_rot".format(t=push_type),
            lambda a, b, c: [b, a, c],
            input_types=[push_type] * 3,
            output_types=[push_type] * 3,
            code_blocks=(3 if push_type == "exec" else 0),
            docstring="Rotates the top three {t}s.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_flush".format(t=push_type),
            _flusher(push_type),
            types_used=[push_type],
            code_blocks=0,
            docstring="Empties the {t} stack.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_eq".format(t=push_type),
            lambda a, b: [a == b],
            input_types=[push_type, push_type],
            output_types=["bool"],
            code_blocks=0,
            docstring="Pushes True if the top two {t} are equal. Otherwise pushes False.".format(t=push_type)
        ))

        i.append(TakesStateInstruction(
            "{t}_stack_depth".format(t=push_type),
            _stack_depther(push_type),
            output_types=["int"],
            other_types=[push_type],
            code_blocks=0,
            docstring="Pushes the size of the {t} stack to the int stack.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_yank".format(t=push_type),
            _yanker(push_type),
            types_used=[push_type, "int"],
            code_blocks=0,
            docstring="Yanks a {t} from deep in the stack based on an index from the int stack and puts it on top.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_yank_dup".format(t=push_type),
            _yank_duper(push_type),
            types_used=[push_type, "int"],
            code_blocks=0,
            docstring="Yanks a copy of a {t} deep in the stack based on an index from the int stack and puts it on top.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_shove".format(t=push_type),
            _shover(push_type),
            types_used=[push_type, "int"],
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Shoves the top {t} deep in the stack based on an index from the int stack.".format(t=push_type)
        ))

        i.append(StateToStateInstruction(
            "{t}_shove_dup".format(t=push_type),
            _shove_duper(push_type),
            types_used=[push_type, "int"],
            code_blocks=(1 if push_type == "exec" else 0),
            docstring="Shoves a copy of the top {t} deep in the stack based on an index from the int stack.".format(t=push_type)
        ))

        i.append(TakesStateInstruction(
            "{t}_is_empty".format(t=push_type),
            _is_emptyer(push_type),
            output_types=["bool"],
            other_types=[push_type],
            code_blocks=0,
            docstring="Pushes True if the {t} stack is empty. Pushes False otherwise.".format(t=push_type)
        ))

    return i



"""The :mod:`search` module defines algorithms to search for Push programs."""
from abc import ABC, abstractmethod
from typing import Union, Tuple
from warnings import warn

from numpy.random import random
import math

from pyshgp.utils import DiscreteProbDistrib
from pyshgp.gp.evaluation import Evaluator
from pyshgp.gp.genome import GeneSpawner, GenomeSimplifier
from pyshgp.gp.individual import Individual
from pyshgp.gp.population import Population
from pyshgp.gp.selection import Selector, get_selector
from pyshgp.gp.variation import VariationOperator, get_variation_operator


# @TODO: Should SearchConfiguration be JSON serializable?
class SearchConfiguration:
    """Configuration of an search algorithm.

    Parameters
    ----------
    evaluator : Evaluator
        The Evaluator to use when evaluating individuals.
    spawning : Union[GeneSpawner, str], optional
        The GeneSpawner, or DiscreteProbDistrib of gene spawners to use when
        producing Genomes during initialization and variation.
    selection :Union[Selector, DiscreteProbDistrib, str], optional
        A Selector, or DiscreteProbDistrib of selectors, to use when selecting
        parents. The default is lexicase selection.
    variation : Union[VariationOperator, DiscreteProbDistrib, str], optional
        A VariationOperator, or DiscreteProbDistrib of VariationOperators, to
        use during variation. Default is SIZE_NEUTRAL_UMAD.
    population_size : int, optional
        The number of individuals hold in the population each generation. Default
        is 300.
    max_generations : int, optional
        The number of generations to run the search algorithm. Default is 100.
    error_threshold : float, optional
        If the search algorithm finds an Individual with a total error less
        than this values, stop searching. Default is 0.0.
    initial_genome_size : Tuple[int, int], optional
        The range of genome sizes to produce during initialization. Default is
        (20, 100)
    simplification_steps : int
        The number of simplification iterations to apply to the best Push program
        produced by the search algorithm.

    """

    def __init__(self,
                 evaluator: Evaluator,
                 spawning: Union[GeneSpawner, DiscreteProbDistrib],
                 selection: Union[Selector, DiscreteProbDistrib, str] = "lexicase",
                 variation: Union[VariationOperator, DiscreteProbDistrib, str] = "umad",
                 population_size: int = 500,
                 max_generations: int = 100,
                 error_threshold: float = 0.0,
                 initial_genome_size: Tuple[int, int] = (10, 50),
                 simplification_steps: int = 2000):
        self.evaluator = evaluator
        self.population_size = population_size
        self.max_generations = max_generations
        self.error_threshold = error_threshold
        self.initial_genome_size = initial_genome_size
        self.simplification_steps = simplification_steps

        if isinstance(spawning, DiscreteProbDistrib):
            self._spawning = spawning
        else:
            self._spawning = DiscreteProbDistrib().add(spawning, 1.0)

        if isinstance(selection, str):
            self._selection = DiscreteProbDistrib().add(get_selector(selection), 1.0)
        elif isinstance(selection, DiscreteProbDistrib):
            self._selection = selection
        else:
            self._selection = DiscreteProbDistrib().add(selection, 1.0)

        if variation == "umad":
            self._variation = DiscreteProbDistrib().add(get_variation_operator, 1.0)
        elif isinstance(variation, DiscreteProbDistrib):
            self._variation = variation
        else:
            self._variation = DiscreteProbDistrib().add(variation, 1.0)

    def get_spawner(self):
        """Return a GeneSpawner."""
        return self._spawning.sample()

    def get_selector(self):
        """Return a Selector."""
        return self._selection.sample()

    def get_variation_operator(self):
        """Return a VariationOperator."""
        return self._variation.sample()


class SearchAlgorithm(ABC):
    """Base class for all search algorithms.

    Parameters
    ----------
    config : SearchConfiguration
        The configuation of the search algorithm.

    Attributes
    ----------
    config : SearchConfiguration
        The configuation of the search algorithm.
    generation : int
        The current generation, or iteration, of the search.
    best_seen : Individual
        The best Individual, with respect to total error, seen so far.
    population : Population
        The current Population of individuals.

    """

    def __init__(self, config: SearchConfiguration):
        self.config = config
        self.generation = 0
        self.best_seen = None
        self.init_population()

    def init_population(self):
        """Initialize the population."""
        self.population = Population()
        for i in range(self.config.population_size):
            spawner = self.config.get_spawner()
            genome = spawner.spawn_genome(self.config.initial_genome_size)
            self.population.add(Individual(genome))

    @abstractmethod
    def step(self) -> bool:
        """Perform one generation (step) of the search. Return if should continue.

        The step method should assume an evaluated Population, and must only
        perform parent selection and variation (producing children). The step
        method should modify the search algorithms population in-place, or
        assign a new Population to the population attribute.

        """
        pass

    def _full_step(self, verbose: bool) -> bool:
        self.generation += 1
        self.population.evaluate(self.config.evaluator)

        best_this_gen = self.population.best()
        if self.best_seen is None or best_this_gen.total_error < self.best_seen.total_error:
            self.best_seen = best_this_gen
            if self.best_seen.total_error <= self.config.error_threshold:
                return False

        if verbose:
            stat_logs = []
            stat_logs.append("GENERATION: {g}".format(
                g=self.generation
            ))
            stat_logs.append("ERRORS: best={b}, median={m}, diversity={e_d}".format(
                b=self.population.best().total_error,
                m=self.population.median_error(),
                e_d=self.population.error_diversity()
            ))
            # stat_logs.append("Population: size={p_s}, genome_diversity={g_d}".format(
            #     p_s=len(self.population),
            #     g_d=self.population.genome_diversity()
            # ))
            print(" | ".join(stat_logs))

        self.step()
        return True

    def run(self, verbose: bool = False):
        """Run the algorithm until termination.

        Parameters
        ----------
        verbose : bool, optional
            Indicates if verbose printing should be used during searching.
            Default is False.

        """
        while self._full_step(verbose):
            if self.generation >= self.config.max_generations:
                break

        simplifier = GenomeSimplifier(self.config.evaluator, verbose)
        simp_genome, simp_error_vector = simplifier.simplify(
            self.best_seen.genome,
            self.best_seen.error_vector,
            self.config.simplification_steps
        )
        simplified_best = Individual(simp_genome)
        simplified_best.error_vector = simp_error_vector

        return simplified_best


class GeneticAlgorithm(SearchAlgorithm):
    """Genetic algorithm to synthesize Push programs.

    An initial Population of random Individuals is created. Each generation
    begins by evaluating all Individuals in the population. Then the current
    Popluation is replaced with children produced by selecting parents from
    the Population and applying VariationOperators to them.

    """

    def __init__(self, config: SearchConfiguration):
        super().__init__(config)

    def _make_child(self) -> Individual:
        op = self.config.get_variation_operator()
        selector = self.config.get_selector()
        parent_genomes = [p.genome for p in selector.select(self.population, n=op.num_parents)]
        child_genome = op.produce(parent_genomes, self.config.get_spawner())
        return Individual(child_genome)

    def step(self):
        """Perform one generation (step) of the genetic algorithm.

        The step method assumes an evaluated Population and performs parent
        selection and variation (producing children).

        """
        self.population = Population(
            [self._make_child() for _ in range(self.config.population_size)]
        )


class SimulatedAnnealing(SearchAlgorithm):
    """Algorithm to synthesize Push programs with Simulated Annealing.

    At each step (generation), the simulated annealing heuristic mutates the
    current Individual, and probabilistically decides between accepting or
    rejecting the child. If the child is accepted, it becomes the new current
    Individual.

    After each step, the simmulated annealing system cools its temperature.
    As the temperature lowers, the probability of accepting a child that does
    not have a lower total error than the current Individual decreases.

    """

    def __init__(self, config: SearchConfiguration):
        if not config.population_size == 1:
            warn("SimulatedAnnealing only supports a population size of 1. Config has been overwritten. ")
            config.population_size = 1

        for op in config._variation.elements:
            assert op.num_parents <= 1, "SimulatedAnnealing cannot take multiple parant variation operators."

        super().__init__(config)

    def _get_temp(self,):
        """Return the temperature."""
        return 1.0 - (self.generation / self.config.max_generations)

    def _acceptance(self, next_error):
        """Return probability of acceptance given an error."""
        current_error = self.population.best().total_error
        if next_error < current_error:
            return 1
        else:
            return math.exp(-(next_error - current_error) / self._get_temp())

    def step(self):
        """Perform one generation, or step, of the Simulated Annealing.

        The step method assumes an evaluated Population one Individual and
        produces a single candidate Individual. If the candidate individual
        passes the acceptance function, it becomes the Individual in the
        Population.

        """
        if self._get_temp() <= 0:
            return

        candidate = Individual(
            self.config.get_variation_operator().produce(
                [self.population.best().genome],
                self.config.get_spawner()
            )
        )
        candidate.error_vector = self.config.evaluator.evaluate(candidate.program)

        acceptance_probability = self._acceptance(candidate.total_error)
        if random() < acceptance_probability:
            self.population = Population().add(candidate)


# @TODO: class EvolutionaryStrategy(SearchAlgorithm):
#     ...


def get_search_algo(name: str) -> SearchAlgorithm:
    """Return the search algorithm class with the given name."""
    name_to_cls = {
        "GA": GeneticAlgorithm,
        "SA": SimulatedAnnealing,
        # "ES": EvolutionaryStrategy,
    }
    search_algo = name_to_cls.get(name, None)
    if search_algo is None:
        raise ValueError("No search algo '{nm}'. Supported names: {lst}.".format(
            nm=name,
            lst=list(name_to_cls.keys())
        ))
    return search_algo



"""Definitions for all core text instructions."""
from pyshgp.push.instruction import SimpleInstruction, ProducesManyOfTypeInstruction
from pyshgp.push.types import Char
from pyshgp.utils import Token


def _concat(a, b):
    return str(b) + str(a),


def _first_char(s):
    if len(s) == 0:
        return Token.revert
    return s[0],


def _last_char(s):
    if len(s) == 0:
        return Token.revert
    return s[-1],


def _nth_char(s, ndx):
    if len(s) == 0:
        return Token.revert
    return s[ndx % len(s)],


def _contains(s, x):
    return x in s,


def _p_index(s, substr):
    try:
        return s.index(substr),
    except ValueError:
        return -1,


def _head(s, i):
    if len(s) == 0:
        return "",
    return s[:i % len(s)],


def _tail(s, i):
    if len(s) == 0:
        return "",
    return s[i % len(s):],


def _rest(s):
    if len(s) < 2:
        return "",
    return s[1:],


def _but_last(s):
    if len(s) < 2:
        return "",
    return s[:-1],


def _drop(s, i):
    if len(s) == 0:
        return "",
    return s[i % len(s):],


def _but_last_n(s, i):
    if len(s) == 0:
        return "",
    return s[:-(i % len(s))],


def _split_on(s, x):
    if x == "":
        return []
    return s.split(x)


def _replace_n(s, old, new, n=1):
    return s.replace(str(old), str(new), n),


def _replace_all(s, old, new):
    return s.replace(str(old), str(new)),


def _remove_n(s, x, n=1):
    return _replace_n(s, x, "", n)


def _remove_all(s, x):
    return _replace_all(s, x, "")


def _len(s):
    return len(s),


def _reverse(s):
    return s[::-1],


def _make_empty():
    return "",


def _is_empty(s):
    return s == "",


def _occurrences_of(s, x):
    return s.count(str(x)),


def _remove_nth(s, ndx):
    return s[:ndx] + s[ndx + 1:],


def _set_nth(s, c, ndx):
    return s[:ndx] + str(c) + s[ndx + 1:],


def _insert(s, x, ndx):
    return s[:ndx] + str(x) + s[ndx:],


def _strip_whitespace(s):
    return s.strip(),


# @TODO: Implement exec_string_iterate instruction.


def _is_whitespace(c):
    return str(c).isspace(),


def _is_letter(c):
    return str(c).isalpha(),


def _is_digit(c):
    return str(c).isdigit(),


def _str_from_thing(thing):
    return str(thing),


def _char_from_bool(b):
    if b:
        return Char("T"),
    return Char("F"),


def _char_from_ascii(i):
    return Char(chr(i % 128)),


def _char_from_float(f):
    return _char_from_ascii(int(f))


def _all_chars(s):
    return [Char(c) for c in list(s)[::-1]]


def instructions():
    """Return all core text instructions."""
    i = []

    for push_type in ["str", "char"]:

        i.append(SimpleInstruction(
            "{t}_concat".format(t=push_type),
            _concat,
            input_types=[push_type, push_type],
            output_types=["str"],
            code_blocks=0,
            docstring="Concatenates the top two {t}s and pushes the resulting string.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "str_insert_{t}".format(t=push_type),
            _insert,
            input_types=["str", push_type, "int"],
            output_types=["str"],
            code_blocks=0,
            docstring="""Inserts {t} into the top str at index `n` and pushes
            the resulting string. The value for `n` is taken from the int stack.""".format(t=push_type)
        ))

        # Getting Characters

        i.append(SimpleInstruction(
            "{t}_from_first_char".format(t=push_type),
            _first_char,
            input_types=["str"],
            output_types=[push_type],
            code_blocks=0,
            docstring="Pushes a {t} of the first character of the top string.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_from_last_char".format(t=push_type),
            _last_char,
            input_types=["str"],
            output_types=[push_type],
            code_blocks=0,
            docstring="Pushes a {t} of the last character of the top string.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "{t}_from_nth_char".format(t=push_type),
            _nth_char,
            input_types=["str", "int"],
            output_types=[push_type],
            code_blocks=0,
            docstring="Pushes a {t} of the nth character of the top string. The top integer denotes nth position.".format(t=push_type)
        ))

        # Checking string contents

        i.append(SimpleInstruction(
            "str_contains_{t}".format(t=push_type),
            _contains,
            input_types=["str", push_type],
            output_types=["bool"],
            code_blocks=0,
            docstring="Pushes true if the next {t} is in the top string. Pushes false otherwise.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "str_index_of_{t}".format(t=push_type),
            _p_index,
            input_types=["str", push_type],
            output_types=["int"],
            code_blocks=0,
            docstring="Pushes the index of the next {t} in the top string. If not found, pushes -1.".format(t=push_type)
        ))

        # Splitting

        i.append(ProducesManyOfTypeInstruction(
            "str_split_on_{t}".format(t=push_type),
            _split_on,
            input_types=["str", push_type],
            output_type="str",
            code_blocks=0,
            docstring="Pushes multiple strs produced by splitting the top str on the top {t}.".format(t=push_type)
        ))

        # Replacements

        i.append(SimpleInstruction(
            "str_replace_first_{t}".format(t=push_type),
            _replace_n,
            input_types=["str", push_type, push_type],
            output_types=["str"],
            code_blocks=0,
            docstring="""Pushes the str produced by replaceing the first occurrence of the
            top {t} with the second {t}.""".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "str_replace_n_{t}".format(t=push_type),
            _replace_n,
            input_types=["str", push_type, push_type, "int"],
            output_types=["str"],
            code_blocks=0,
            docstring="""Pushes the str produced by replaceing the first `n` occurrences of the
            top {t} with the second {t}. The value for `n` is the top int.""".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "str_replace_all_{t}".format(t=push_type),
            _replace_all,
            input_types=["str", push_type, push_type],
            output_types=["str"],
            code_blocks=0,
            docstring="""Pushes the str produced by replaceing all occurrences of the
            top {t} with the second {t}.""".format(t=push_type)
        ))

        # Removals

        i.append(SimpleInstruction(
            "str_remove_first_{t}".format(t=push_type),
            _remove_n,
            input_types=["str", push_type],
            output_types=["str"],
            code_blocks=0,
            docstring="Pushes the str produced by removing the first occurrence of the top {t}.".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "str_remove_n_{t}".format(t=push_type),
            _remove_n,
            input_types=["str", push_type, "int"],
            output_types=["str"],
            code_blocks=0,
            docstring="""Pushes the str produced by remvoing the first `n` occurrences of the
            top {t}. The value for `n` is the top int.""".format(t=push_type)
        ))

        i.append(SimpleInstruction(
            "str_remove_all_{t}".format(t=push_type),
            _remove_all,
            input_types=["str", push_type],
            output_types=["str"],
            code_blocks=0,
            docstring="Pushes the str produced by removing all occurrences of the top {t}.".format(t=push_type)
        ))

        # Misc

        i.append(SimpleInstruction(
            "str_occurrences_of_{t}".format(t=push_type),
            _occurrences_of,
            input_types=["str", push_type],
            output_types=["int"],
            code_blocks=0,
            docstring="Pushes the number of times the top {t} occurs in the top str to the int stack.".format(t=push_type)
        ))

    i.append(SimpleInstruction(
        "str_reverse",
        _reverse,
        input_types=["str"],
        output_types=["str"],
        code_blocks=0,
        docstring="""Takes the top string and pushes it reversed."""
    ))

    i.append(SimpleInstruction(
        "str_head",
        _head,
        input_types=["str", "int"],
        output_types=["str"],
        code_blocks=0,
        docstring="""Pushes a string of the first `n` characters from the top string. The value
        for `n` is the top int mod the length of the string."""
    ))

    i.append(SimpleInstruction(
        "str_tail",
        _tail,
        input_types=["str", "int"],
        output_types=["str"],
        code_blocks=0,
        docstring="""Pushes a string of the last `n` characters from the top string. The value
        for `n` is the top int mod the length of the string."""
    ))

    i.append(SimpleInstruction(
        "str_append_char",
        _concat,
        input_types=["str", "char"],
        output_types=["str"],
        code_blocks=0,
        docstring="Appends the top char to the top string pushes the resulting string."
    ))

    i.append(SimpleInstruction(
        "str_rest",
        _rest,
        input_types=["str"],
        output_types=["str"],
        code_blocks=0,
        docstring="Pushes the top str without its first character."
    ))

    i.append(SimpleInstruction(
        "str_but_last",
        _but_last,
        input_types=["str"],
        output_types=["str"],
        code_blocks=0,
        docstring="Pushes the top str without its last character."
    ))

    i.append(SimpleInstruction(
        "str_drop",
        _drop,
        input_types=["str", "int"],
        output_types=["str"],
        code_blocks=0,
        docstring="""Pushes the top str without its first `n` character. The value for `n`
        is the top int mod the length of the string."""
    ))

    i.append(SimpleInstruction(
        "str_but_last_n",
        _but_last_n,
        input_types=["str", "int"],
        output_types=["str"],
        code_blocks=0,
        docstring="""Pushes the top str without its last `n` character. The value for `n`
        is the top int mod the length of the string."""
    ))

    i.append(SimpleInstruction(
        "str_length",
        _len,
        input_types=["str"],
        output_types=["int"],
        code_blocks=0,
        docstring="Pushes the length of the top str to the int stack."
    ))

    i.append(SimpleInstruction(
        "str_make_empty",
        _make_empty,
        input_types=[],
        output_types=["str"],
        code_blocks=0,
        docstring="Pushes an empty string."
    ))

    i.append(SimpleInstruction(
        "str_is_empty_string",
        _is_empty,
        input_types=["str"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Pushes True if top string is empty. Pushes False otherwise."
    ))

    i.append(SimpleInstruction(
        "str_remove_nth",
        _remove_nth,
        input_types=["str", "int"],
        output_types=["str"],
        code_blocks=0,
        docstring="Pushes the top str with the nth character removed."
    ))

    i.append(SimpleInstruction(
        "str_set_nth",
        _set_nth,
        input_types=["str", "char", "int"],
        output_types=["str"],
        code_blocks=0,
        docstring="Pushes the top str with the nth character set to the top character."
    ))

    i.append(SimpleInstruction(
        "str_strip_whitespace",
        _strip_whitespace,
        input_types=["str"],
        output_types=["str"],
        code_blocks=0,
        docstring="Pushes the top str with trailing and leading whitespace stripped."
    ))

    # @TODO: Instructions for trim_left and trim_right
    # @TODO: Instructions for pad_left and pad_right

    #  CHARACTER INSTRUCTIONS

    i.append(SimpleInstruction(
        "char_is_whitespace",
        _is_whitespace,
        input_types=["char"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Pushes True if the top Char is whitespace. Pushes False otherwise."
    ))

    i.append(SimpleInstruction(
        "char_is_letter",
        _is_letter,
        input_types=["char"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Pushes True if the top Char is a letter. Pushes False otherwise."
    ))

    i.append(SimpleInstruction(
        "char_is_digit",
        _is_digit,
        input_types=["char"],
        output_types=["bool"],
        code_blocks=0,
        docstring="Pushes True if the top Char is a numeric digit. Pushes False otherwise."
    ))

    #  TYPE CONVERTING

    for push_type in ["bool", "int", "float", "char"]:
        i.append(SimpleInstruction(
            "str_from_{t}".format(t=push_type),
            _str_from_thing,
            input_types=[push_type],
            output_types=["str"],
            code_blocks=0,
            docstring="Pushes the top {t} converted into a str.".format(t=push_type)
        ))

    i.append(SimpleInstruction(
        "char_from_bool",
        _char_from_bool,
        input_types=["bool"],
        output_types=["char"],
        code_blocks=0,
        docstring="""Pushes the char \"T\" if the top bool is True. If the top
        bool is False, pushes the char \"F\"."""
    ))

    i.append(SimpleInstruction(
        "char_from_ascii_int",
        _char_from_ascii,
        input_types=["int"],
        output_types=["char"],
        code_blocks=0,
        docstring="Pushes the top int converted into a Character by using the int mod 128 as an ascii value."
    ))

    i.append(SimpleInstruction(
        "_char_from_float",
        _char_from_float,
        input_types=["float"],
        output_types=["char"],
        code_blocks=0,
        docstring="""Pushes the top float converted into a Character by flooring
        the float to an int, taking the int mod 128, and using it as an ascii value."""
    ))

    i.append(ProducesManyOfTypeInstruction(
        "chars_from_str",
        _all_chars,
        input_types=["str"],
        output_type="char",
        code_blocks=0,
        docstring="""Pushes each character of the top str to the char stack in reverse order."""
    ))

    return i



"""Definitions for all core I/O instructions, including input instructions."""
from typing import Sequence, Callable

from pyshgp.push.state import PushState
from pyshgp.push.atoms import Literal
from pyshgp.push.instruction import SimpleInstruction, TakesStateInstruction
from pyshgp.push.types import PUSH_TYPES


def _nth_inputer(ndx: int) -> Callable:
    # @TODO: Replace with partial
    def f(state: PushState) -> Sequence[Literal]:
        return [Literal(state.inputs[ndx])]
    return f


def make_input_instruction(ndx: int) -> TakesStateInstruction:
    """Return insctuction to push a copy of the input value at the given index."""
    return TakesStateInstruction(
        "input_{i}".format(i=ndx),
        _nth_inputer(ndx),
        output_types=["exec"],
        other_types=[],
        code_blocks=0,
        docstring="Push a copy of input at index {i}.".format(i=ndx)
    )


def make_input_instructions(num_inputs: int) -> Sequence[TakesStateInstruction]:
    """Return insctuctions to push a copy of the input value at the given index."""
    return [make_input_instruction(i) for i in range(num_inputs)]


# Printing instructions

def instructions():
    """Return all core printing instructions."""
    i = []

    for push_type in PUSH_TYPES:
        i.append(SimpleInstruction(
            "print_{t}".format(t=push_type.name),
            lambda x: [str(x)],
            input_types=[push_type.name],
            output_types=["stdout"],
            code_blocks=0,
            docstring="Prints the top {t}.".format(t=push_type.name)
        ))
    return i



"""The state module is define the PushState class.

A PushState object holds the PushStacks, stdout string, and collection of input
values. The PushState is what controls the setup of all stacks before program
manipulation, and the producing of outputs after program execution.
"""
from typing import Sequence, Union, Any, Set
import numpy as np

from pyshgp.push.types import PushType, push_type_by_name
from pyshgp.push.atoms import Atom, CodeBlock
from pyshgp.push.stack import PushStack
from pyshgp.utils import Token


class PushState(dict):
    """A collection of PushStacks used during push program execution."""

    __slots__ = ["stdout", "inputs", "_jit_push_types"]

    def __init__(self, push_types: Set[Union[PushType, str]]):
        super().__init__()
        self.stdout: str = ""
        self.inputs: Sequence[Any] = []
        self._jit_push_types: dict = {}

        for push_type in set(list(push_types) + ["exec"]):
            if isinstance(push_type, str):
                push_type_from_name = push_type_by_name(push_type)
                if push_type_from_name is None:
                    push_type = PushType(push_type, (Atom,))
                    push_type.coerce = Atom.coerce
                    self._jit_push_types[push_type.name] = push_type
                else:
                    push_type = push_type_from_name
            self._register_stack_of_type(push_type)

    def __eq__(self, other) -> bool:
        if not isinstance(other, PushState):
            return False
        return super().__eq__(other) and self.inputs == other.inputs and self.stdout == other.stdout

    def _register_stack_of_type(self, push_type: PushType):
        """Add a new stack."""
        self[push_type.name] = PushStack(push_type)

    @classmethod
    def from_dict(cls, d):
        """Set the state to match the given dictionary.

        .. warning::
            This is written to be used in ``pyshgp`` tests, NOT as part of
            push program execution or evolution. There are no checks to confirm
            that the ``d`` can be converted to a valid Push state.

        Parameters
        ----------
        d : dict
            Dict that is converted into a Push state.

        """
        state = cls(d.keys())
        inputs = []
        stdout = ""
        for k in d.keys():
            if k == 'inputs':
                inputs = d[k]
            elif k == 'stdout':
                stdout += d[k]
            else:
                for v in d[k]:
                    state[k].push(v)

        state.load_inputs(inputs)
        state.stdout = stdout
        return state

    def load_program(self, program: CodeBlock):
        """Push the given CodeBlock to the execution stack."""
        self["exec"].push(program)

    def load_inputs(self, inputs):
        """Load a list of input values onto the PushState inputs.

        Parameters
        ----------
        inputs : list
            List of input values.

        """
        if not isinstance(inputs, (list, np.ndarray)):
            raise ValueError(
                "Push inputs must be a list, got {t}".format(t=type(inputs)))
        self.inputs = inputs

    def observe_stacks(self, types: Sequence[str]) -> list:
        """Return a list of output values based on the given types indicated.

        Items are take from the tops of each stack. If multiple occurences of
        the same type are in ``output_types``, the returned values are taken
        from progressively deeper in that stack. Does not pop the values off
        the stacks.

        Parameters
        ----------
        types : list
            List of strings denoting the push types of the returned values.

        """
        values = []
        counts = {}
        for typ in types:
            ndx = counts.get(typ, 0)
            values.append(self[typ].nth(ndx))
            counts[typ] = ndx + 1
        return values

    def pop_from_stacks(self, types: Sequence[str]) -> Union[Sequence, Token]:
        """Pop the top items for each value_type. Return a vector of the values popped from each stack."""
        values = []
        for typ in types:
            val = self[typ].top()
            if val is Token.no_stack_item:
                return Token.revert
            else:
                values.append(val)
                self[typ].pop()
        return values

    def push_to_stacks(self, values: list, types: Sequence[str]):
        """Check that all values can be coerced into their expected PushType. Push them onto the correct stack."""
        for ndx in range(len(values)):
            val = values[ndx]
            typ = types[ndx]
            if typ == "stdout":
                self.stdout += str(val)
            else:
                self[typ].push(val)

    def size(self):
        """Return the size of the PushState."""
        return sum([len(s) for s in self.values()]) + len(self.inputs)

    def pretty_print(self):
        """Print the state of all stacks in the PushState."""
        for k, v in self.items():
            print(k, ":", v)
        print('inputs :', self.inputs)
        print('stdout :', self.stdout)



"""Utility classes and functions used throughout pyshgp."""
from typing import Optional, Union
from abc import ABC, abstractmethod
from enum import Enum

import numpy as np
from numpy.random import choice


class DiscreteProbDistrib:
    """Discrete Probability Distribution."""

    __slots__ = ["elements", "_total", "_raw_probabilities", "_normalized_probabilities"]

    def __init__(self):
        self.elements = []
        self._total = 0.0
        self._raw_probabilities = []
        self._normalize()

    def _normalize(self):
        self._normalized_probabilities = np.array(self._raw_probabilities) / self._total

    def add(self, el, p):
        """Add an element with a relative probability."""
        self.elements.append(el)
        self._total += float(p)
        self._raw_probabilities.append(p)
        self._normalize()
        return self

    def size(self):
        """Return the number of elements in the distribution."""
        return len(self.elements)

    def sample(self):
        """Return a sample from the distribution."""
        if self.size() == 1:
            return self.elements[0]
        return choice(self.elements, p=self._normalized_probabilities)

    def sample_n(self, n: int = 1, replace: bool = True):
        """Return n samples from the distribution."""
        if self.size() == 1 and replace:
            return [self.elements[0]] * n
        return choice(self.elements, n, replace, self._normalized_probabilities)


class Token(Enum):
    """Enum class of all Tokens."""

    no_stack_item = 1
    revert = 2
    whole_state = 3


class PushError(Exception):
    """Error raised during Push program execution."""

    @classmethod
    def no_type(cls, thing):
        """Raise PushError when no PushType can be found for something."""
        return cls("Unkown PushType for {th}.".format(th=thing))

    @classmethod
    def failed_coerce(cls, thing, push_type):
        """Raise PushError when something fails to coerce to a PushType."""
        return cls("Could not convert {typ1} {th} to {typ2}.".format(
            th=thing,
            typ1=type(thing),
            typ2=push_type
        ))

    @classmethod
    def empty_character(cls):
        """Raise PushError when Character is made from empty string."""
        return cls("Character object cannot be created from empty string.")

    @classmethod
    def long_character(cls):
        """Raise PushError when Character is made from string length more than 1."""
        return cls("Character object cannot be created from string of length > 1.")


class JSONable(ABC):
    """Abstract base class for objects can be transformed into JSON."""

    @abstractmethod
    def jsonify(self) -> str:
        """Return the object as a JSON string."""
        pass

    def to_json(self, filepath: Optional[str] = None) -> Optional[str]:
        """Write the object to either a string or a file."""
        json_str = self.jsonify()
        if filepath is None:
            return json_str
        else:
            with open(filepath, "w+") as f:
                f.write(json_str)


def jsonify_collection(root: Union[list, dict]) -> str:
    """Return the given list or dict and all elements as a JSON string."""
    def _helper(thing) -> str:
        if isinstance(thing, list):
            return "[" + ",".join([_helper(el) for el in thing]) + "]"
        elif isinstance(thing, dict):
            return "{" + ",".join([str(k) + ":" + _helper(v) for k, v in thing.items()]) + "}"
        elif isinstance(thing, JSONable):
            return thing.jsonify()
        else:
            return str(thing)

    if isinstance(root, (list, dict)):
        return _helper(root)
    else:
        raise ValueError("Can jsonify_collection lists and dicts. Got {t}.".format(t=type(root)))



"""The :mod:`variation` module defines classes for variation operators.

Variation operators (aka genetic operators) are used in evolutionary/genetic
algorithms to create "child" genomes from "parent" genomes.

"""
from abc import ABC, abstractmethod
from typing import Sequence, Union
import math
from copy import copy

from numpy.random import random, choice

from pyshgp.push.types import PushType
from pyshgp.push.atoms import Literal
from pyshgp.gp.genome import Genome, GeneSpawner
from pyshgp.utils import DiscreteProbDistrib


class VariationOperator(ABC):
    """Base class of all VariationOperators.

    Parameters
    ----------
    num_parents : int
        Number of parent Genomes the operator needs to produce a child
        Individual.

    Attributes
    ----------
    num_parents : int
        Number of parent Genomes the operator needs to produce a child
        Individual.

    """

    def __init__(self, num_parents: int):
        self.num_parents = num_parents

    def checknum_parents(self, parents: Sequence[Genome]):
        """Raise error if given too few parents.

        Parameters
        ----------
        parents
            A list of parent Genomes given to the operator.

        """
        if not len(parents) >= self.num_parents:
            raise ValueError("Variation operator given {a} parents. Expected {e}.".format(
                a=len(parents),
                e=self.num_parents)
            )

    @abstractmethod
    def produce(self, parents: Sequence[Genome], spawner: GeneSpawner = None) -> Genome:
        """Produce a child Genome from parent Genomes and optional GenomeSpawner.

        Parameters
        ----------
        parents
            A list of parent Genomes given to the operator.
        spawner
            A GeneSpawner that can be used to produce new genes (aka Atoms).

        """
        pass


class VariationStrategy(DiscreteProbDistrib):
    """A collection of VariationOperator and how frequently to use them."""

    def add(self, op: VariationOperator, p: float):
        """Add an element with a relative probability.

        Parameters
        ----------
        op : VariationOperator
            The VariationOperator to add to the variaiton strategy.
        p : float
            The probability of using the given operator relative to the other
            operators that have been added to the VariationStrategy.

        """
        super().add(op, p)


class VariationPipeline(VariationOperator):
    """Variation operator that sequencially applies multiple others variation operators.

    Parameters
    ----------
    operators : list of VariationOperators
        A list of operators to apply in order to produce the child Genome.

    Attributes
    ----------
    operators : list of VariationOperators
        A list of operators to apply in order to produce the child Genome.
    num_parents : int
        Number of parent Genomes the operator needs to produce a child
        Individual.

    """

    def __init__(self, operators: Sequence[VariationOperator]):
        num_parents_needed = max([op.num_parents for op in operators])
        super().__init__(num_parents_needed)
        self.operators = operators

    def produce(self, parents: Sequence[Genome], spawner: GeneSpawner) -> Genome:
        """Produce a child Genome from parent Genomes and optional GenomeSpawner.

        Parameters
        ----------
        parents
            A list of parent Genomes given to the operator.
        spawner
            A GeneSpawner that can be used to produce new genes (aka Atoms).

        """
        self.checknum_parents(parents)
        child = parents[0].copy()
        for op in self.operators:
            child = op.produce([child] + parents[1:], spawner)
        return child


# Utilities

def _gaussian_noise_factor():
    """Return Gaussian noise of mean 0, std dev 1.

    Returns
    --------
    Float samples from Gaussian distribution.

    Examples
    --------
    >>> gaussian_noise_factor()
    1.43412557975
    >>> gaussian_noise_factor()
    -0.0410900866765

    """
    return math.sqrt(-2.0 * math.log(random())) * math.cos(2.0 * math.pi * random())


# Mutations

# @TODO: Implement all the common literal mutations.
class LiteralMutation(VariationOperator, ABC):
    """Base class for mutations of literal Atoms.

    Parameters
    ----------
    push_type : pyshgp.push.types.PushType
        The PushType which the operator can mutate.
    rate : float
        The probablility of applying the mutation to a given Literal.

    Attributes
    ----------
    push_type : pyshgp.push.types.PushType
        The PushType which the operator can mutate.
    rate : float
        The probablility of applying the mutation to a given Literal.
    num_parents : int
        Number of parent Genomes the operator needs to produce a child
        Individual.

    """

    def __init__(self, push_type: PushType, rate: float = 0.01):
        super().__init__(1)
        self.rate = rate
        self.push_type = push_type

    @abstractmethod
    def _mutate_literal(literal: Literal) -> Literal:
        ...

    def produce(self, parents: Sequence[Genome], spawner: GeneSpawner = None) -> Genome:
        """Produce a child Genome from parent Genomes and optional GenomeSpawner.

        Parameters
        ----------
        parents
            A list of parent Genomes given to the operator.
        spawner
            A GeneSpawner that can be used to produce new genes (aka Atoms).

        """
        self.checknum_parents(parents)
        new_genome = Genome()
        for atom in parents[0]:
            if isinstance(atom, Literal) and self.push_type == atom.push_type and random() < self.rate:
                new_atom = self._mutate_literal(atom)
            else:
                new_atom = atom
            new_genome.append(new_atom)
        return new_genome


class DeletionMutation(VariationOperator):
    """Uniformly randomly removes some Atoms from parent.

    Parameters
    ----------
    rate : float
        The probablility of removing any given Atom in the parent Genome.
        Default is 0.01.

    Attributes
    ----------
    rate : float
        The probablility of removing any given Atom in the parent Genome.
        Default is 0.01.
    num_parents : int
        Number of parent Genomes the operator needs to produce a child
        Individual.

    """

    def __init__(self, deletion_rate: float = 0.01):
        super().__init__(1)
        self.rate = deletion_rate

    def produce(self, parents: Sequence[Genome], spawner: GeneSpawner) -> Genome:
        """Produce a child Genome from parent Genomes and optional GenomeSpawner.

        Parameters
        ----------
        parents
            A list of parent Genomes given to the operator.
        spawner
            A GeneSpawner that can be used to produce new genes (aka Atoms).

        """
        self.checknum_parents(parents)
        new_genome = Genome()
        for gene in parents[0]:
            if random() < self.rate:
                continue
            new_genome.append(gene)
        return new_genome


class AdditionMutation(VariationOperator):
    """Uniformly randomly adds some Atoms to parent.

    Parameters
    ----------
    rate : float
        The probablility of adding a new Atom at any given point in the parent
        Genome. Default is 0.01.

    Attributes
    ----------
    rate : float
        The probablility of adding a new Atom at any given point in the parent
        Genome. Default is 0.01.
    num_parents : int
        Number of parent Genomes the operator needs to produce a child
        Individual.

    """

    def __init__(self, addition_rate: float = 0.01):
        super().__init__(1)
        self.rate = addition_rate

    def produce(self, parents: Sequence[Genome], spawner: GeneSpawner) -> Genome:
        """Produce a child Genome from parent Genomes and optional GenomeSpawner.

        Parameters
        ----------
        parents
            A list of parent Genomes given to the operator.
        spawner
            A GeneSpawner that can be used to produce new genes (aka Atoms).

        """
        self.checknum_parents(parents)
        new_genome = Genome()
        for gene in parents[0]:
            if random() < self.rate:
                new_genome.append(spawner.spawn_atom())
            new_genome.append(gene)
        return new_genome


# Recombinations

class Alternation(VariationOperator):
    """Uniformly alternates between the two parent genomes.

    Parameters
    ----------
    rate : float, optional (default=0.01)
        The probablility of switching which parent program elements are being
        copied from. Must be 0 <= rate <= 1. Defaults to 0.1.
    alignment_deviation : int, optional (default=10)
        The standard deviation of how far alternation may jump between indices
        when switching between parents.

    Attributes
    ----------
    rate : float, optional (default=0.01)
        The probablility of switching which parent program elements are being
        copied from. Must be 0 <= rate <= 1. Defaults to 0.1.
    alignment_deviation : int, optional (default=10)
        The standard deviation of how far alternation may jump between indices
        when switching between parents.
    num_parents : int
        Number of parent Genomes the operator needs to produce a child
        Individual.

    """

    def __init__(self, alternation_rate=0.01, alignment_deviation=10):
        super().__init__(2)
        self.rate = alternation_rate
        self.alignment_deviation = alignment_deviation

    def produce(self, parents: Sequence[Genome], spawner: GeneSpawner = None) -> Genome:
        """Produce a child Genome from parent Genomes and optional GenomeSpawner.

        Parameters
        ----------
        parents
            A list of parent Genomes given to the operator.
        spawner
            A GeneSpawner that can be used to produce new genes (aka Atoms).

        """
        self.checknum_parents(parents)
        gn1 = parents[0].copy()
        gn2 = parents[1].copy()
        new_genome = Genome()
        # Random pick which parent to start from
        use_parent_1 = choice([True, False])
        loop_times = len(gn1)
        if not use_parent_1:
            loop_times = len(gn2)
        i = 0
        while (i < loop_times):
            if random() < self.rate:
                # Switch which parent we are pulling genes from
                i += round(self.alignment_deviation * _gaussian_noise_factor())
                i = int(max(0, i))
                use_parent_1 = not use_parent_1
            else:
                # Pull gene from parent
                if use_parent_1:
                    new_genome.append(gn1[i])
                else:
                    new_genome.append(gn2[i])
                i = int(i + 1)
            # Change loop stop condition
            loop_times = len(gn1)
            if not use_parent_1:
                loop_times = len(gn2)
        return new_genome


# Other

class Genesis(VariationOperator):
    """Creates an entirely new (and random) genome.

    Parameters
    ----------
    size
        The child genome will contain this many Atoms if size is an integer.
        If size is a pair of integers, the genome will be of a random
        size in the range of the two integers.

    Attributes
    ----------
    size
        The child genome will contain this many Atoms if size is an integer.
        If size is a pair of integers, the genome will be of a random
        size in the range of the two integers.
    num_parents : int
        Number of parent Genomes the operator needs to produce a child
        Individual.

    """

    def __init__(self, *, size: Union[int, Sequence[int]]):
        super().__init__(0)
        self.size = size

    def produce(self, parents: Sequence[Genome], spawner: GeneSpawner) -> Genome:
        """Produce a child Genome from parent Genomes and optional GenomeSpawner.

        Parameters
        ----------
        parents
            A list of parent Genomes given to the operator.
        spawner
            A GeneSpawner that can be used to produce new genes (aka Atoms).

        """
        return spawner.spawn_genome(self.size)


class Cloning(VariationOperator):
    """Clones the parent genome.

    Attributes
    ----------
    num_parents : int
        Number of parent Genomes the operator needs to produce a child
        Individual.

    """

    def __init__(self):
        super().__init__(1)

    def produce(self, parents: Sequence[Genome], spawner: GeneSpawner = None) -> Genome:
        """Produce a child Genome from parent Genomes and optional GenomeSpawner.

        Parameters
        ----------
        parents
            A list of parent Genomes given to the operator.
        spawner
            A GeneSpawner that can be used to produce new genes (aka Atoms).

        """
        return copy.copy(parents[0])


def get_variation_operator(name: str) -> VariationOperator:
    """Get the variaton operator class with the given name."""
    name_to_cls = {
        "deletion": DeletionMutation,
        "addition": AdditionMutation,
        "alternation": Alternation,
        # UMAD citation: https://dl.acm.org/citation.cfm?id=3205455.3205603
        "umad": VariationPipeline([AdditionMutation(0.09), DeletionMutation(0.0826)]),
        "umad-shrink": VariationPipeline([AdditionMutation(0.09), DeletionMutation(0.1)]),
        "umad-grow": VariationPipeline([AdditionMutation(0.09), DeletionMutation(0.0652)])
    }
    op = name_to_cls.get(name, None)
    if op is None:
        raise ValueError("No varition operator '{nm}'. Supported names: {lst}.".format(
            nm=name,
            lst=list(name_to_cls.keys())
        ))
    return op



# -*- coding: utf-8 -*-
"""The :mod:`stack` module defines the ``PushStack`` class.

A PushStack is used to hold values of a certain PushType in a PushState object.
"""
from typing import Optional

from pyshgp.push.types import PushType
from pyshgp.utils import Token


class PushStack(list):
    """Stack that holds elements of a sinlge PushType.

    Parameters
    ----------
    push_type : PushType
        The PushType all items of the stack should conform to, or be coerced into.

    Attributes
    ----------
    push_type : PushType
        The PushType all items of the stack should conform to, or be coerced into.

    """

    __slots__ = ["push_type"]

    def __init__(self, push_type: PushType):
        self.push_type = push_type

    def is_empty(self) -> bool:
        """Return True if the stack is empty. Return False otherwise."""
        return len(self) == 0

    def push(self, value):
        """Pushes a value to the top of the stack.

        An alias for `append` which makes the definition of push instructions
        more clear.

        Parameters
        ----------
        value :
            Value to push onto stack.

        """
        if not self.push_type.is_instance(value):
            value = self.push_type.coerce(value)
        self.append(value)
        return self

    def pop(self, index: Optional[int] = None):
        """Pop the top item off the stack.

        Parameters
        ----------
        index :
            Index to pop from the stack.

        Returns
        --------
        Element at ``index`` in stack.

        """
        if index is None:
            return super().pop()
        else:
            return super().pop((len(self) - 1) - index)

    def nth(self, position: int):
        """Return the element at a given position.

        If stack is empty, returns None. If ``position < 0`` or
        ``position > len(self)`` returns a ``no_stack_item`` token.

        Parameters
        ----------
        position : int
            Position in stack to get item.

        Returns
        --------
        Element at ``positon`` in stack.

        """
        if len(self) <= position:
            return Token.no_stack_item
        elif len(self) == 0:
            return Token.no_stack_item
        else:
            return self[(len(self) - 1) - position]

    def take(self, n: int):
        """Return the top `n` items from the tack.

        If `n` is less than zero, result is an empty list. If `n` is greater
        than the size of the stack, result contains every element.

        Parameters
        ----------
        n : int
            Number of items

        Returns
        --------
        Element at ``positon`` in stack.

        """
        if n < 0:
            return []
        return self[:-(n + 1):-1]

    def top(self):
        """Return the top item on the stack, or a ``no_stack_item`` token if empty.

        Returns
        --------
        Returns the top element of the stack, or None if empty.

        """
        if len(self) > 0:
            return self[-1]
        else:
            return Token.no_stack_item

    def insert(self, position: int, value):
        """Insert value at position in stack.

        Parameters
        ----------
        position : int
            Position in stack to get item.
        value :
            Value to insert into stack.

        """
        if not self.push_type.is_instance(value):
            value = self.push_type.coerce(value)
        super().insert(len(self) - position, value)
        return self

    def set_nth(self, position: int, value):
        """Overwrite the item in the nth position of the stack with the new value.

        Parameters
        ----------
        position : int
            Position in stack to get item.
        value :
            Value to insert into stack.

        """
        if not self.push_type.is_instance(value):
            value = self.push_type.coerce(value)
        self[len(self) - 1 - position] = value
        return self

    def flush(self):
        """Empty the stack."""
        del self[:]
        return self

    def __repr__(self):
        self_r = list(self)[::-1]
        return self_r.__repr__()

    def __eq__(self, other):
        if not isinstance(other, PushStack):
            return False
        return self.push_type == other.push_type and list(self) == list(other)



"""The :mod:`estimator` module defines a ``PushEstimator`` class."""
import json
import inspect
from typing import Union, Tuple, Sequence

import numpy as np
import pandas as pd
from sklearn.utils.validation import check_is_fitted, check_X_y

import pyshgp.gp.search as sr
import pyshgp.gp.selection as sl
import pyshgp.gp.variation as vr
from pyshgp.gp.evaluation import DatasetEvaluator
from pyshgp.gp.genome import GeneSpawner
from pyshgp.push.instruction_set import InstructionSet
from pyshgp.push.interpreter import PushInterpreter, DEFAULT_INTERPRETER
from pyshgp.push.atoms import CodeBlock
from pyshgp.push.types import push_type_for_type
from pyshgp.utils import JSONable


class SearchResult(JSONable):
    """The best program found by a search algorith and the types it outputs.

    Parameters
    ----------
    program : CodeBlock
        The Push program found by a SearchAlgorithm.
    output_types : Sequence[str]
        The types output by the program.

    Attributes
    ----------
    program : CodeBlock
        The Push program found by a SearchAlgorithm.
    output_types : Sequence[str]
        The types output by the program.

    """

    def __init__(self, program: CodeBlock, output_types: Sequence[str]):
        self.program = program
        self.output_types = output_types

    @classmethod
    def from_json_str(cls, json_str: str, instruction_set: InstructionSet):
        """Create a PushSolution from a JSON string."""
        solution_dict = json.loads(json_str)
        return cls(
            CodeBlock.from_json_str(
                json.dumps(solution_dict["program"], separators=(',', ':')),
                instruction_set
            ),
            solution_dict["output_types"]
        )

    @classmethod
    def from_json_file(cls, filepath: str, instruction_set: InstructionSet):
        """Create a PushSolution from a JSON string."""
        with open(filepath, "r") as f:
            return cls.from_json_str(f.read(), instruction_set)

    def jsonify(self):
        """Return the object as a JSON string."""
        return '{{"program":{p},"output_types":{o}}}'.format(
            p=self.program.to_json(),
            o="[" + ",".join(['"' + t + '"' for t in self.output_types]) + "]"
        )


class PushEstimator:
    """Simple estimator that synthesizes Push programs.

    Parameters
    ----------
    search : Union[SearchAlgorithm, str], optional
        The search algorithm, or its abbreviation, to use to when synthesizing
        Push programs.
    spawner : Union[GeneSpawner, str], optional
        The GeneSpawner to use when producing Genomes during initialization and
        variation. Default is all core intructions, no literals, and no ERC Generators.
    selector : Union[Selector, str], optional
        The selector, or name of selector, to use when selecting parents.
        The default is lexicase selection.
    variation_strategy : Union[VariationStrategy, dict, str]
        A VariationStrategy describing a collection of VariationOperators and how
        frequently to use them. If a dict is supplied, keys should be operator
        names and values should be the probability distirution. If a string is
        provided, the VariationOperators with that name will always be used.
    population_size : int, optional
        The number of individuals hold in the population each generation. Default
        is 300.
    max_generations : int, optional
        The number of generations to run the search algorithm. Default is 100.
    initial_genome_size : Tuple[int, int], optional
        The range of genome sizes to produce during initialization. Default is
        (20, 100)
    simplification_steps : int
        The number of simplification iterations to apply to the best Push program
        produced by the search algorithm.
    interpreter : PushInterpreter, optional
        The PushInterpreter to use when making predictions. Also holds the instruction
        set to use
    **kwargs
        Arbitrary keyword arguments. Examples of supported arguments are
        `epsilon` (bool or float) when using Lexicase as the selector, and
        `tournament_size` (int) when using tournament selection.

    """

    def __init__(self,
                 search: Union[sr.SearchAlgorithm, str] = "GA",
                 spawner: Union[GeneSpawner, str] = "default",
                 selector: Union[sl.Selector, str] = "lexicase",
                 variation_strategy: Union[vr.VariationStrategy, dict, str] = "umad",
                 population_size: int = 300,
                 max_generations: int = 100,
                 initial_genome_size: Tuple[int, int] = (20, 100),
                 simplification_steps: int = 2000,
                 interpreter: PushInterpreter = "default",
                 **kwargs):
        self.search = search
        self.spawner = spawner
        self.selector = selector
        self.variation_strategy = variation_strategy
        self.population_size = population_size
        self.max_generations = max_generations
        self.initial_genome_size = initial_genome_size
        self.simplification_steps = simplification_steps

        if interpreter == "default":
            self.interpreter = DEFAULT_INTERPRETER
        else:
            self.interpreter = interpreter

        self.ext = kwargs

    def _build_component(self, component_cls, **kwargs):
        arg_names = inspect.getfullargspec(component_cls.__init__)[0][1:]
        args = kwargs
        for arg_name in arg_names:
            if arg_name in self.ext:
                args[arg_name] = self.ext[arg_name]
        return component_cls(**args)

    def _build_search_algo(self):
        if isinstance(self.spawner, GeneSpawner):
            self._spawner = self.spawner
        elif self.spawner == "default":
            self._spawner = GeneSpawner(
                instruction_set=self.interpreter.instruction_set,
                literals=[],
                erc_generators=[]
            )
        else:
            raise ValueError("Bad spawner: {}".format(self.spawner))

        if isinstance(self.selector, sl.Selector):
            self._selector = self.selector
        else:
            selector_cls = sl.get_selector(self.selector)
            self._selector = self._build_component(selector_cls)

        if isinstance(self.variation_strategy, vr.VariationStrategy):
            self._variation_strategy = self.variation_strategy
        else:
            self._variation_strategy = vr.VariationStrategy()
            if isinstance(self.variation_strategy, dict):
                for op_name, prob in self.variation_strategy.items():
                    var_op = vr.get_variation_operator(op_name)
                    if not isinstance(var_op, vr.VariationOperator):
                        var_op = self._build_component(var_op)
                    self._variation_strategy.add(var_op, prob)
            else:
                var_op = vr.get_variation_operator(self.variation_strategy)
                if not isinstance(var_op, vr.VariationOperator):
                    var_op = self._build_component(var_op)
                self._variation_strategy.add(var_op, 1.0)

        search_config = sr.SearchConfiguration(
            spawning=self._spawner,
            evaluator=self.evaluator,
            selection=self._selector,
            variation=self._variation_strategy,
            population_size=self.population_size,
            max_generations=self.max_generations,
            initial_genome_size=self.initial_genome_size,
            simplification_steps=self.simplification_steps
        )

        if isinstance(self.search, sr.SearchAlgorithm):
            self._search = self.search
        else:
            search_algo_cls = sr.get_search_algo(self.search)
            self._search = self._build_component(search_algo_cls, config=search_config)

    def fit(self, X, y, verbose: bool = False):
        """Run the search algorithm to synthesize a push program.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The training input samples.
        y : list, array-like, or pandas dataframe.
            The target values (class labels in classification, real numbers in
            regression). Shape = [n_samples] or [n_samples, n_outputs]
        verbose : bool, optional
            Indicates if verbose printing should be used during searching.
            Default is False.

        """
        X, y = check_X_y(X, y, force_all_finite=False, allow_nd=True, multi_output=True)

        if isinstance(X, (pd.DataFrame, np.ndarray)):
            if len(X.shape) > 1:
                arity = X.shape[1]
            else:
                arity = 1
        elif isinstance(X, (pd.Series, list)):
            try:
                arity = len(X[0])
            except TypeError:
                arity = 1
        else:
            arity = 1
        self.interpreter.instruction_set.register_n_inputs(arity)

        if isinstance(y, pd.DataFrame):
            y_types = list(y.dtypes)
        elif isinstance(y, pd.Series):
            y_types = [y.dtype]
        elif isinstance(y, np.ndarray):
            if len(y.shape) > 1:
                y_types = [y.dtype] * y.shape[1]
            else:
                y_types = [y.dtype]
        elif isinstance(y[0], list):
            # @TODO: Implement a better way to detect type than check the first element.
            y_types = [type(_) for _ in y[0]]
        else:
            y_types = [type(y[0])]
        output_types = [push_type_for_type(t).name for t in y_types]

        self.evaluator = DatasetEvaluator(X, y, interpreter=self.interpreter)
        self._build_search_algo()
        best_seen = self._search.run(verbose)

        self._result = SearchResult(best_seen.program, output_types)

    def predict(self, X, verbose: bool = False):
        """Execute the synthesized push program on a dataset.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The set of cases to predict.
        verbose : bool, optional
            Indicates if verbose printing should be used during searching.
            Default is False.

        Returns
        -------
        y_hat : pandas dataframe of shape = [n_samples, n_outputs]

        """
        check_is_fitted(self, "_result")
        return [
            self.interpreter.run(self._result.program, inputs, self._result.output_types, verbose=verbose) for inputs in X
        ]

    def score(self, X, y):
        """Run the search algorithm to synthesize a push program.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The training input samples.

        y : list, array-like, or pandas dataframe.
            The target values (class labels in classification, real numbers in
            regression). Shape = [n_samples] or [n_samples, n_outputs]

        """
        check_is_fitted(self, "_result")
        self.evaluator = DatasetEvaluator(X, y)
        return self.evaluator.evaluate(self._result.program)

    def save(self, filepath: str):
        """Load the found solution to a JSON file.

        Parameters
        ----------
        filepath
            Filepath to write the serialized search result to.

        """
        check_is_fitted(self, "_result")
        self._result.to_json(filepath)

    def load(self, filepath: str):
        """Load a found solution from a JSON file.

        Parameters
        ----------
        filepath
            Filepath to read the serialized search result from.

        """
        self._result = SearchResult.from_json_file(filepath, self.interpreter.instruction_set)



"""The :mod:`selection` module defines classes to select Individuals from Populations."""
from abc import ABC, abstractmethod
from typing import Sequence, Union
from operator import attrgetter

import numpy as np
from numpy.random import random, choice, shuffle

from pyshgp.gp.individual import Individual
from pyshgp.gp.population import Population


class Selector(ABC):
    """Base class for all selection algorithms."""

    @abstractmethod
    def select_one(self, population: Population) -> Individual:
        """Return single individual from population.

        Parameters
        ----------
        population
            A Population of Individuals.

        Returns
        -------
        Individual
            The selected Individual.

        """
        pass

    def select(self, population: Population, n: int = 1) -> Sequence[Individual]:
        """Return `n` individuals from the population.

        Parameters
        ----------
        population
            A Population of Individuals.

        Returns
        -------
        Sequence[Individual]
            The selected Individuals.

        """
        selected = []
        for i in range(n):
            selected.append(self.select_one(population))
        return selected


# @TODO: class SelectorPipeline(Selector)


class FitnessProportionate(Selector):
    """Fitness proportionate selection, also known as roulette wheel selection.

    See: https://en.wikipedia.org/wiki/Fitness_proportionate_selection
    """

    def select_one(self, population: Population) -> Individual:
        """Return single individual from population.

        Parameters
        ----------
        population
            A Population of Individuals.

        Returns
        -------
        Individual
            The selected Individual.

        """
        return self.select(population)[0]

    def select(self, population: Population, n: int = 1) -> Sequence[Individual]:
        """Return `n` individuals from the population.

        Parameters
        ----------
        population
            A Population of Individuals.

        Returns
        -------
        Sequence[Individual]
            The selected Individuals.

        """
        population_total_errors = np.array([i.total_error for i in population])
        sum_of_total_errors = np.sum(population_total_errors)
        probablilities = 1.0 - (population_total_errors / sum_of_total_errors)
        selected_ndxs = np.searchsorted(np.cumsum(probablilities), random(n))
        return [population[ndx] for ndx in selected_ndxs]


class Tournament(Selector):
    """Tournament selection.

    See: https://en.wikipedia.org/wiki/Tournament_selection

    Parameters
    ----------
    tournament_size : int, optional
        Number of individuals selected uniformly randomly to participate in
        the tournament. Default is 7.

    Attributes
    ----------
    tournament_size : int, optional
        Number of individuals selected uniformly randomly to participate in
        the tournament. Default is 7.

    """

    def __init__(self, tournament_size: int = 7):
        self.tournament_size = tournament_size

    def select_one(self, population: Population) -> Individual:
        """Return single individual from population.

        Parameters
        ----------
        population
            A Population of Individuals.

        Returns
        -------
        Individual
            The selected Individual.

        """
        tournament = choice(population, self.tournament_size, replace=False)
        return min(tournament, key=attrgetter('total_error'))


def median_absolute_deviation(x: np.ndarray) -> np.float64:
    """Return the MAD.

    Parameters
    ----------
    a : array-like, shape = (n,)

    Returns
    -------
    mad : float

    """
    return np.median(np.abs(x - np.median(x)))


class Lexicase(Selector):
    """Lexicase Selection.

    All training cases are considered iteratively in a random order. For each
    training cases, the population is filtered to only contain the Individuals
    which have an error value within epsilon of the best error value on that case.
    This filtering is repeated until the population is down to a single Individual
    or all cases have been used. After the filtering iterations, a random
    Individual from the remaining set is returned as the selected Individual.

    See: https://ieeexplore.ieee.org/document/6920034
    """

    def __init__(self, epsilon: Union[bool, float, np.ndarray] = False):
        self.epsilon = epsilon

    def _preselection(self, population: Population) -> Sequence[Individual]:
        """Preselect one individual per distinct error vector.

        Crucial for avoiding the worst case runtime of lexicase selection but
        does not impact the behavior of which indiviudal gets selected.
        """
        population_list = list(population)
        shuffle(population_list)
        preselected = []
        error_vector_hashes = []
        for individual in population_list:
            error_vector_hash = hash(individual.error_vector_bytes)
            if error_vector_hash not in error_vector_hashes:
                preselected.append(individual)
                error_vector_hashes.append(error_vector_hash)
        return preselected

    def select_one(self, population: Population) -> Individual:
        """Return single individual from population.

        Parameters
        ----------
        population
            A Population of Individuals.

        Returns
        -------
        Individual
            The selected Individual.

        """
        candidates = self._preselection(population)
        cases = np.arange(len(population[0].error_vector))
        shuffle(cases)

        if isinstance(self.epsilon, np.ndarray):
            ep = np.apply_along_axis(median_absolute_deviation, 0, population.all_error_vectors())
        elif isinstance(self.epsilon, (float, int, np.int64, np.float64)):
            ep = self.epsilon

        while len(cases) > 0 and len(candidates) > 1:
            case = cases[0]
            errors_this_case = [i._error_vector[case] for i in candidates]
            best_val_for_case = min(errors_this_case)

            if isinstance(self.epsilon, np.ndarray):
                max_error = best_val_for_case + ep[case]
            elif isinstance(self.epsilon, (float, int, np.int64, np.float64)):
                max_error = best_val_for_case + ep
            else:
                max_error = best_val_for_case

            candidates = [i for i in candidates if i._error_vector[case] <= max_error]
            cases = cases[1:]
        return choice(candidates)


class Elite(Selector):
    """Returns the best N individuals by total error."""

    def select_one(self, population: Population) -> Individual:
        """Return single individual from population.

        Parameters
        ----------
        population
            A Population of Individuals.

        Returns
        -------
        Individual
            The selected Individual.

        """
        return population.best()

    def select(self, population: Population, n: int = 1) -> Sequence[Individual]:
        """Return `n` individuals from the population.

        Parameters
        ----------
        population
            A Population of Individuals.

        Returns
        -------
        Sequence[Individual]
            The selected Individuals.

        """
        return population.best_n(n)


def get_selector(name: str) -> Selector:
    """Get the selector class with the given name."""
    name_to_cls = {
        "roulette": FitnessProportionate,
        "tournament": Tournament,
        "lexicase": Lexicase,
        "elite": Elite,
    }
    selector = name_to_cls.get(name, None)
    if selector is None:
        raise ValueError("No selector '{nm}'. Supported names: {lst}.".format(
            nm=name,
            lst=list(name_to_cls.keys())
        ))
    return selector



"""The ``types`` module contains the core PushTypes and functions to reference them.

A PushType simply is a named collection of Python and Numpy types. A PushType
can be used to determine if multiple items should be considered the same type
during Push program execution.

"""

from typing import Any, Tuple, Optional

import numpy as np

from pyshgp.utils import PushError


class PushType:
    """Type information for values used by Push programs.

    Parameters
    ----------
    name : str
        A name for the type. Used when referencing the PushType in Instruction
        definitions and will be the key in the PushState for the corresponding
        PushStack.
    underlying : Tuple[type]
        A tuple of python (or numpy) types that correspond to the underlying
        native types which the PushType is representing.

    Attributes
    ----------
    name : str
        A name for the type. Used when referencing the PushType in Instruction
        definitions and will be the key in the PushState for the corresponding
        PushStack.
    underlying : Tuple[type]
        A tuple of python (or numpy) types that correspond to the underlying
        native types which the PushType is representing.

    """

    def __init__(self, name: str, underlying: Tuple[type]):
        self.name = name
        self.underlying = underlying

    def is_instance(self, thing: Any) -> bool:
        """Return true if thing is instance of PushTypes underlying type(s)."""
        return isinstance(thing, self.underlying)

    def coerce(self, thing: Any):
        """Convert thing into PushTypes underlying type."""
        try:
            return self.underlying[0](thing)
        except ValueError:
            raise PushError.failed_coerce(thing, self)

    def __repr__(self):
        return self.name + "<" + ",".join([t.__name__ for t in list(self.underlying)]) + ">"

    def __eq__(self, other):
        if not isinstance(other, PushType):
            return False
        return self.name == other.name and self.underlying == other.underlying

    def __hash__(self):
        return (self.name + str(self.underlying[0])).__hash__()


class _PushIntType(PushType):

    def __init__(self):
        super().__init__("int", (int, np.int32, np.int64))


class _PushFloatType(PushType):

    def __init__(self):
        super().__init__("float", (float, np.float32, np.float64))


class _PushStrType(PushType):

    def __init__(self):
        super().__init__("str", (str, np.str_, np.object, np.dtype('<U28')))


class _PushBoolType(PushType):

    def __init__(self):
        super().__init__("bool", (bool, np.bool_,))


class Char(str):
    """Holds a string of length 1.

    Used to distinguish between string and char literals in Push program
    interpretation.

    Attributes
    ----------
    char : str
        String of length 1.

    """

    def __init__(self, char):
        if len(char) == 0:
            raise PushError.empty_character()
        if len(char) > 1:
            raise PushError.long_character()

    def __eq__(self, other):
        return isinstance(other, Char) and str(self) == str(other)


class _PushCharType(PushType):
    """The Character PushType."""

    def __init__(self):
        super().__init__("char", (Char, ))


# @TODO: Add vector type(s)


PushInt = _PushIntType()
PushFloat = _PushFloatType()
PushStr = _PushStrType()
PushBool = _PushBoolType()
PushChar = _PushCharType()

PUSH_TYPES = [PushBool, PushInt, PushChar, PushFloat, PushStr]

_type_lookup = dict([(t.name, t) for t in PUSH_TYPES])


def push_type_by_name(push_type_name: str) -> Optional[PushType]:
    """Return the PushType with given name. If not exits, return None."""
    return _type_lookup.get(push_type_name)


def push_type_of(thing: Any) -> Optional[PushType]:
    """Return the PushType of the given thing."""
    for push_type in PUSH_TYPES:
        if push_type.is_instance(thing):
            return push_type
    return None


def push_type_for_type(t: type) -> Optional[PushType]:
    """Return the PushType of the given python (or numpy) type."""
    for push_type in PUSH_TYPES:
        if t in push_type.underlying:
            return push_type
    return None



"""The :mod:`genome` module defines classes related to Genomesself.

The ``Genome`` class defines Genomes as flat, linear representations of Push
programs. The ``GenomeSpawner`` class is a factory of random genes (``Atoms``)
and random ``Genomes``.

"""
from typing import Callable, Sequence, Union, Tuple, Any
from copy import copy, deepcopy

import numpy as np

from pyshgp.push.atoms import (
    Atom, Closer, Literal, Instruction, CodeBlock, AtomFactory
)
from pyshgp.push.instruction_set import InstructionSet
from pyshgp.gp.evaluation import Evaluator
from pyshgp.utils import DiscreteProbDistrib, JSONable, jsonify_collection


class Opener:
    """Marks the start of one or more CodeBlock."""

    __slots__ = ["count"]

    def __init__(self, count: int):
        self.count = count

    def dec(self):
        """Decrements the count by 1."""
        self.count -= 1


def _has_opener(l: Sequence) -> bool:
    return sum([isinstance(_, Opener) for _ in l]) > 0


class Genome(list, Atom, JSONable):
    """A flat sequence of Atoms where each Atom is a "gene" in the genome."""

    def __init__(self, atoms: Sequence[Atom] = None):
        if atoms is not None:
            for el in atoms:
                self.append(el)

    def append(self, el: Atom):
        """Append Atom to end of Genome."""
        if isinstance(el, CodeBlock):
            raise ValueError("Cannot add CodeBlock to genomes. Genomes must be kept flat.")
        super().append(el)

    @staticmethod
    def from_json_str(json_str: str, instruction_set: InstructionSet):
        """Create a Genome from a JSON string."""
        atoms = AtomFactory.json_str_to_atom_list(json_str, instruction_set)
        return Genome(atoms)

    def to_code_block(self) -> CodeBlock:
        """Translate into nested CodeBlocks.

        These CodeBlocks can be considered the Push program representation of
        the Genome which can be executed by a PushInterpreter and evaluated
        by an Evaluator.

        """
        plushy_buffer = []
        for atom in self:
            plushy_buffer.append(atom)
            if isinstance(atom, Instruction) and atom.code_blocks > 0:
                plushy_buffer.append(Opener(atom.code_blocks))

        push_buffer = []
        while True:
            # If done with plush but unclosed opens, recur with one more close.
            if len(plushy_buffer) == 0 and _has_opener(push_buffer):
                plushy_buffer.append(Closer())
            # If done with plush and all opens closed, return push.
            elif len(plushy_buffer) == 0:
                return CodeBlock.from_list(push_buffer)
            else:
                atom = plushy_buffer[0]
                # If next instruction is a close, and there is an open.
                if isinstance(atom, Closer) and _has_opener(push_buffer):
                    ndx, opener = [(ndx, el) for ndx, el in enumerate(push_buffer) if isinstance(el, Opener)][-1]
                    post_open = push_buffer[ndx + 1:]
                    pre_open = push_buffer[:ndx]
                    if opener.count == 1:
                        push_buffer = pre_open + [post_open]
                    else:
                        opener.dec()
                        push_buffer = pre_open + [post_open, opener]
                # If next instruction is a close, and there is no open.
                elif not isinstance(atom, Closer):
                    push_buffer.append(atom)
                del plushy_buffer[0]

    def simplify(self,
                 evaluator: Evaluator,
                 origional_error: float,
                 steps: int,
                 verbose: bool = False):
        """Simplify genome while preserving or improving error."""
        simplifier = GenomeSimplifier(evaluator, verbose)
        gn, _ = simplifier.simplify(self, origional_error, steps)
        return gn

    def copy(self, *, deep: bool = False):
        """Return a copy of the Genome."""
        if deep:
            return deepcopy(self)
        return copy(self)

    def jsonify(self) -> str:
        """Return the object as a JSON string."""
        return jsonify_collection(self)


class GeneSpawner:
    """A factory of random Genes (Atoms) and Genomes.

    When  spawning a random gene, the result can be one of three types of Atoms.
    An Instruction, a Closer, or a Literal. If the Atom is a Literal, it may
    be one of the supplied Literals, or it may be the result of running one of
    the Ephemeral Random Constant generators.

    Reference for ERCs:
    "A field guide to genetic programming", Section 3.1
    Riccardo Poli and William B. Langdon and Nicholas Freitag McPhee,
    http://www.gp-field-guide.org.uk/

    Parameters
    ----------
    instruction_set : pyshgp.push.instruction_set.InstructionSet
        InstructionSet containing instructions to use when spawning genes and
        genomes.
    literals : Sequence[pyshgp.push.instruction_set.atoms.Literal, Any]
        A list of Literal objects to pull from when spawning genes and genomes.
        If an element of ``literals`` is not an instance of Literal, an attempt
        to wrap the value in a Literal will be made.
    erc_generator : Sequence[Callable]
        A list of functions (aka Ephemeral Random Constant generators). When one
        of these functions is called, the output is placed in a Literal and
        returned as the spawned gene.
    distribution : pyshgp.utils.DiscreteProbDistrib, optional
        A probability distribution describing how frequently to produce
        Instructions, Closers, Literals, and ERCs. The default is "proportional"
        which gives all Instructions, Literals, and ERC generators equal probability.
        If "proportional", the likelyhood of producing a Closer is the same as
        the likelyhood of producing an Instruction with opens a code block.

    Attributes
    ----------
    instruction_set : pyshgp.push.instruction_set.InstructionSet
        InstructionSet containing instructions to use when spawning genes and
        genomes.
    literals : Sequence[pyshgp.push.instruction_set.atoms.Literal]
        A list of Literal objects to pull from when spawning genes and genomes.
    erc_generator : Sequence[Callable]
        A list of functions (aka Ephemeral Random Constant generators). When one
        of these functions is called, the output is placed in a Literal and
        returned as the spawned gene.
    distribution : pyshgp.utils.DiscreteProbDistrib
        A probability distribution describing how frequently to produce
        Instructions, Closers, Literals, and ERCs.

    """

    def __init__(self,
                 instruction_set: InstructionSet,
                 literals: Sequence[Union[Literal, Any]],
                 erc_generators: Sequence[Callable],
                 distribution: DiscreteProbDistrib = "proportional"):
        self.instruction_set = instruction_set
        self.literals = [lit if isinstance(lit, Literal) else Literal(lit) for lit in literals]
        self.erc_generators = erc_generators

        if distribution == "proportional":
            self.distribution = (
                DiscreteProbDistrib()
                .add("instruction", len(instruction_set))
                .add("close", sum([i.code_blocks for i in instruction_set.values()]))
                .add("literal", len(literals))
                .add("erc", len(erc_generators))
            )
        else:
            self.distribution = distribution

    def random_instruction(self) -> Instruction:
        """Return a random Instruction from the InstructionSet.

        Returns
        -------
        pushgp.push.atoms.Instruction
            A randomly selected Literal.

        """
        return np.random.choice(list(self.instruction_set.values()))

    def random_literal(self) -> Literal:
        """Return a random Literal from the set of Literals.

        Returns
        -------
        pushgp.push.atoms.Literal
            A randomly selected Literal.

        """
        return np.random.choice(self.literals)

    def random_erc(self) -> Literal:
        """Materialize a random ERC generator into a Literal and return it.

        Returns
        -------
        pushgp.push.atoms.Literal
            A Literal whose value comes from running a ERC generator function.

        """
        erc_value = np.random.choice(self.erc_generators)()
        return Literal(erc_value)

    def spawn_atom(self) -> Atom:
        """Return a random Atom based on the GenomeSpawner's distribution.

        Returns
        -------
        pushgp.push.atoms.Atom
            An random Atom. Either an Instruction, Closer, or Literal.

        """
        atom_type = self.distribution.sample()
        if atom_type == "instruction":
            return self.random_instruction()
        elif atom_type == "close":
            return Closer()
        elif atom_type == "literal":
            return self.random_literal()
        elif atom_type == "erc":
            return self.random_erc()
        else:
            raise ValueError("GenomeSpawner distribution bad atom type {t}".format(t=str(atom_type)))

    def spawn_genome(self, size: Union[int, Sequence[int]]) -> Genome:
        """Return a random Genome based on the GenomeSpawner's distribution.

        The genome will contain the specified number of Atoms if size is an
        integer. If size is a pair of integers, the genome will be of a random
        size in the range of the two integers.

        Parameters
        ----------
        size
            The resulting genome will contain this many Atoms if size is an
            integer. If size is a pair of integers, the genome will be of a random
            size in the range of the two integers.

        Returns
        -------
        pushgp.gp.genome.Genome
            A Genome with random contents of a given size.

        """
        if isinstance(size, Sequence):
            size = np.random.randint(size[0], size[1]) + 1

        gn = Genome()
        for ndx in range(size):
            gn.append(self.spawn_atom())

        return gn


class GenomeSimplifier:
    """Simplifies a genome while preserving, or improving, its error.

    Genomes, and Push programs, can contain superfluous Push code. This extra
    code often has no effect on the program behavior, but occasionally it can
    introduce subtle errors or behaviors that is not covered by the training
    cases. Removing the superfluous code makes genomes (and thus programs)
    smaller and easier to understand. More importantly, simplification can
    imporve the generalization of the given genome/program.

    The process of geneome simplification is iterative and closely resembles
    simple hill climbing. For each iteration, the simplifier will randomly
    select a small number of random genes to remove. The Genome is re-evaluated
    and if its error gets worse, the change is reverted. After repeating this
    for some number of steps, the resulting genome will be the same size or
    smaller while containing the same (or better) error value.

    Reference:
    "Improving generalization of evolved programs through automatic simplification"
    Thomas Helmuth, Nicholas Freitag McPhee, Edward Pantridge, and Lee Spector. 2017.
    In Proceedings of the Genetic and Evolutionary Computation Conference (GECCO '17).
    ACM, New York, NY, USA, 937-944. DOI: https://doi.org/10.1145/3071178.3071330

    https://dl.acm.org/citation.cfm?id=3071178.3071330

    """

    # @TODO: Add noop swaps to simplification.

    def __init__(self, evaluator: Evaluator, verbose: bool = False):
        self.evaluator = evaluator
        self.verbose = verbose

    def _remove_rand_genes(self, genome: Genome) -> Genome:
        gn = genome.copy()
        n_genes_to_remove = min(np.random.randint(1, 4), len(genome) - 1)
        ndx_of_genes_to_remove = np.random.choice(np.arange(len(gn)), n_genes_to_remove, replace=False)
        ndx_of_genes_to_remove[::-1].sort()
        for ndx in ndx_of_genes_to_remove:
            del gn[ndx]
        return gn

    def _errors_of_genome(self, genome: Genome) -> float:
        program = genome.to_code_block()
        return self.evaluator.evaluate(program)

    def _step(self, genome: Genome, errors_to_beat: np.ndarray) -> Tuple[Genome, float]:
        new_gn = self._remove_rand_genes(genome)
        new_errs = self._errors_of_genome(new_gn)
        if np.sum(new_errs) <= np.sum(errors_to_beat):
            if self.verbose:
                print("Simplified genome to length {ln}.".format(ln=len(new_gn)))
            return new_gn, new_errs
        return genome, errors_to_beat

    def simplify(self,
                 genome: Genome,
                 origional_errors: np.ndarray,
                 steps: int = 2000) -> Tuple[Genome, np.ndarray]:
        """Simplify the given geome while maintaining error.

        Parameters
        ----------
        genome
            The Genome to simplifiy.
        origional_errors
            Error vector of the genome to simplify.
        steps
            Number of simplification iterations to perform. Default is 2000.

        Returns
        -------
        pushgp.gp.genome.Genome
            A Genome with random contents of a given size.

        """
        print("Simplifying genome of length {ln}.".format(ln=len(genome)))
        gn = genome
        errs = origional_errors
        for step in range(steps):
            gn, errs = self._step(gn, errs)
            if len(gn) == 1:
                break
        return gn, errs



"""The :mod:`individual` module defines an Individaul in an evolutionary population.

Individuals are made up of Genomes, which are the linear Push program
representations which can be manipulated by seach algorithms.

"""
from typing import Union

import numpy as np

from pyshgp.push.atoms import CodeBlock
from pyshgp.gp.genome import Genome


class Individual:
    """An individual in an evolutionary population.

    Attributes
    ----------
    genome : Genome
        The Genome of the Individual.
    program : CodeBlock
        The CodeBlock produced by translating the Individual's genome.
    error_vector : np.array
        An array of error values produced by evaluating the Individual's program.
    total_error : float
        The sum of all error values in the Individaul's error_vector.
    error_vector_bytes:
        Hashable Byte representation of the individual's error vector.

    """

    __slots__ = ["_genome", "_program", "_error_vector", "_total_error", "_error_vector_bytes"]

    def __init__(self, genome: Genome = Genome):
        self._genome = genome
        self._program = None
        self._error_vector = None
        self._total_error = None
        self._error_vector_bytes = None

    @property
    def genome(self) -> Genome:
        """Plush Genome of individual."""
        return self._genome

    @genome.setter
    def genome(self, value: Genome):
        self._genome = value

    @property
    def program(self) -> CodeBlock:
        """Push program of individual. Taken from Plush genome."""
        if self._program is None:
            self._program = self.genome.to_code_block()
        return self._program

    @program.setter
    def program(self, value: CodeBlock):
        raise AttributeError("Cannot set program directly. Must set genome.")

    @property
    def error_vector(self) -> np.ndarray:
        """Numpy array of numeric error values."""
        return self._error_vector

    @error_vector.setter
    def error_vector(self, error_vector: np.ndarray):
        self._error_vector = error_vector

    @property
    def total_error(self) -> Union[np.int64, np.float64]:
        """Numeric sum of the error vector."""
        if self._total_error is None:
            try:
                self._total_error = np.sum(self.error_vector)
            except OverflowError:
                self._total_error = np.inf
        return self._total_error

    @total_error.setter
    def total_error(self, value: Union[np.int64, np.float64]):
        raise AttributeError("Cannot set total_error directly. Must set error_vector.")

    @property
    def error_vector_bytes(self):
        """Hashable Byte representation of the individual's error vector."""
        if self._error_vector_bytes is None:
            self._error_vector_bytes = self._error_vector.data.tobytes()
        return self._error_vector_bytes

    @error_vector_bytes.setter
    def error_vector_bytes(self, value):
        raise AttributeError("Cannot set error_vector_bytes directly. Must set error_vector.")

    def __lt__(self, other):
        return self.total_error < other.total_error



"""The :mod:`population` module defines an evolutionary population of Individuals."""
from collections.abc import Sequence
from bisect import insort_left

import numpy as np

from pyshgp.gp.individual import Individual
from pyshgp.gp.evaluation import Evaluator


class Population(Sequence):
    """A sequence of Individuals kept in sorted order, with respect to their total errors."""

    __slots__ = ["unevaluated", "evaluated"]

    def __init__(self, individuals: list = None):
        self.unevaluated = []
        self.evaluated = []

        if individuals is not None:
            for el in individuals:
                self.add(el)

    def __len__(self):
        return len(self.evaluated) + len(self.unevaluated)

    def __getitem__(self, key: int) -> Individual:
        if key < len(self.evaluated):
            return self.evaluated[key]
        return self.unevaluated[key - len(self.evaluated)]

    def add(self, individual: Individual):
        """Add an Individaul to the population."""
        if individual.total_error is None:
            self.unevaluated.append(individual)
        else:
            insort_left(self.evaluated, individual)
        return self

    def best(self):
        """Return the best n individual in the population."""
        return self.evaluated[0]

    def best_n(self, n: int):
        """Return the best n individuals in the population."""
        return self.evaluated[:n]

    def evaluate(self, evaluator: Evaluator):
        """Evaluate all unevaluated individuals in the population."""
        for individual in self.unevaluated:
            individual.error_vector = evaluator.evaluate(individual.program)
            insort_left(self.evaluated, individual)
        self.unevaluated = []

    def all_error_vectors(self):
        """2D array containing all Individuals' error vectors."""
        return np.array([i.error_vector for i in self.evaluated])

    def all_total_errors(self):
        """1D array containing all Individuals' total errors."""
        return np.array([i.total_error for i in self.evaluated])

    def median_error(self):
        """Median total error in the population."""
        return np.median(self.all_total_errors())

    def error_diversity(self):
        """Proportion of unique error vectors."""
        return len(np.unique(self.all_error_vectors(), axis=0)) / float(len(self))

    def genome_diversity(self):
        """Proportion of unique genomes."""
        unq = set([i.genome.jsonify() for i in self])
        return len(unq) / float(len(self))



"""The :mod:`evaluation` module defines classes to evaluate program CodeBlocks."""
from abc import ABC, abstractmethod
from typing import Sequence, Union, Callable
from collections import defaultdict
import numpy as np

from pyshgp.push.interpreter import PushInterpreter, DEFAULT_INTERPRETER
from pyshgp.push.atoms import CodeBlock
from pyshgp.push.types import push_type_of
from pyshgp.utils import Token


def damerau_levenshtein_distance(a: Union[str, Sequence], b: Union[str, Sequence]) -> int:
    """Damerau Levenshtein Distance that works for both strings and lists.

    https://en.wikipedia.org/wiki/Damerau%E2%80%93Levenshtein_distance.
    This implemenation is heavily inspired by the implementation in the
    jellyfish package. https://github.com/jamesturk/jellyfish
    """
    a_is_str = isinstance(a, str)
    b_is_str = isinstance(b, str)
    if a_is_str or b_is_str:
        assert a_is_str and b_is_str

    len1 = len(a)
    len2 = len(b)
    infinite = len1 + len2

    da = defaultdict(int)
    score = [[0] * (len2 + 2) for x in range(len1 + 2)]
    score[0][0] = infinite
    for i in range(0, len1 + 1):
        score[i + 1][0] = infinite
        score[i + 1][1] = i
    for i in range(0, len2 + 1):
        score[0][i + 1] = infinite
        score[1][i + 1] = i

    for i in range(1, len1 + 1):
        db = 0
        for j in range(1, len2 + 1):
            i1 = da[b[j - 1]]
            j1 = db
            cost = 1
            if a[i - 1] == b[j - 1]:
                cost = 0
                db = j

            score[i + 1][j + 1] = min(score[i][j] + cost,
                                      score[i + 1][j] + 1,
                                      score[i][j + 1] + 1,
                                      score[i1][j1] + (i - i1 - 1) + 1 + (j - j1 - 1))
        da[a[i - 1]] = i
    return score[len1 + 1][len2 + 1]


class Evaluator(ABC):
    """Base class or evaluators.

    Parameters
    ----------
    interpreter : PushInterpreter, optional
        PushInterpreter used to run program and get their output. Default is
        an interpreter with the default configuration and all core instructions
        registered.
    penalty : float, optional
        When a program's output cannot be evaluated on a particular case, the
        penalty error is assigned. Default is 5e5.
    verbose : bool, optional
        Indicates if verbose printing should be used during evalutation.
        Default is False.

    """

    def __init__(self, interpreter: PushInterpreter = "default",
                 penalty: float = np.inf, verbose: bool = False):
        self.penalty = penalty
        self.verbose = verbose
        if interpreter == "default":
            self.interpreter = DEFAULT_INTERPRETER
        else:
            self.interpreter = interpreter

    def default_error_function(self, actuals, expecteds) -> np.array:
        """Produce errors of actual program output given expected program output.

        The default error function is intented to be a universal error function
        for Push programs which only output a subset of the standard data types.

        Parameters
        ----------
        actuals : list
            The values produced by running a Push program on a sequences of cases.
        expecteds: list
            The ground truth values for the sequence of cases used to produce the actuals.

        Returns
        -------
        np.array
            An array of error values describing the program's performance.

        """
        errors = []
        for ndx, actual in enumerate(actuals):
            expected = expecteds[ndx]
            if actual is Token.no_stack_item:
                errors.append(self.penalty)
            elif isinstance(expected, (bool, np.bool_)):
                errors.append(int(not (bool(actual) == expected)))
            elif isinstance(expected, (int, np.int64, float, np.float64)):
                try:
                    errors.append(abs(float(actual) - expected))
                except OverflowError:
                    errors.append(self.penalty)
            elif isinstance(expected, str):
                errors.append(damerau_levenshtein_distance(str(actual), expected))
            elif isinstance(expected, list):
                errors += list(self.default_error_function(list(actual), expected))
            else:
                raise ValueError("Unknown expected type for {e}".format(e=expected))
        return np.array(errors)

    @abstractmethod
    def evaluate(self, program: CodeBlock) -> np.ndarray:
        """Evaluate the program and return the error vector.

        Parameters
        ----------
        program
            Program (CodeBlock of Push code) to evaluate.

        Returns
        -------
        np.ndarray
            The error vector of the program.

        """
        pass


class DatasetEvaluator(Evaluator):
    """Evaluator driven by a labeled dataset."""

    def __init__(self, X, y, interpreter: PushInterpreter = "default",
                 penalty: float = np.inf, verbose: bool = False):
        """Create Evaluator based on a labeled dataset. Inspired by sklearn.

        Parameters
        ----------
        X : pandas dataframe of shape = [n_samples, n_features]
            The inputs to evaluate each program on.

        y : list, array-like, or pandas dataframe.
            The target values. Shape = [n_samples] or [n_samples, n_outputs]

        interpreter : PushInterpreter or {"default"}
            The interpreter used to run the push programs.

        penalty : float
            If no response is given by the program on a given input, assign this
            error as the error.

        verbose : bool (default = False)
            Print all program traces while evaluating.

        """
        super().__init__(interpreter, penalty, verbose=verbose)
        self.X = X
        self.y = y

    def evaluate(self, program: CodeBlock) -> np.ndarray:
        """Evaluate the program and return the error vector.

        Parameters
        ----------
        program
            Program (CodeBlock of Push code) to evaluate.

        Returns
        -------
        np.ndarray
            The error vector of the program.

        """
        errors_on_cases = []
        for ndx, case in enumerate(self.X):
            expected = self.y[ndx]
            if not isinstance(expected, (list, np.ndarray)):
                expected = [expected]
            output_types = [push_type_of(_).name for _ in expected]
            actual = self.interpreter.run(program, case, output_types, verbose=self.verbose)
            errors_on_cases.append(self.default_error_function(actual, expected))
        return np.array(errors_on_cases).flatten()


class FunctionEvaluator(Evaluator):
    """Evaluator driven by an error function."""

    def __init__(self, error_function: Callable):
        """Create Evaluator driven by an error function.

        The given error function must take a push program in the form of a
        CodeBlock and then return an np.ndarray of numeric errors. These errors
        will be used as the program's error vector.

        The error functions will typically instansiate its own PushInterpreter
        an run the given program as needed.

        Parameters
        ----------
        error_function : Callable
            A function which takes a program to evaluate and returns a
            np.ndarray of errors.

        """
        super().__init__()
        self.error_function = error_function

    def evaluate(self, program: CodeBlock) -> np.ndarray:
        """Evaluate the program and return the error vector.

        Parameters
        ----------
        program
            Program (CodeBlock of Push code) to evaluate.

        Returns
        -------
        np.ndarray
            The error vector of the program.

        """
        return self.error_function(program)



"""Pyshgp setup file."""
import os
from setuptools import setup, find_packages


exec(open("pyshgp/__init__.py").read())


def read(fname):
    """Read a file to a string."""
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="pyshgp",
    version=__version__,
    description="Push Genetic Programming in Python",
    long_description=read('README.md'),
    long_description_content_type="text/markdown",
    keywords=["push gp", "genetic programming", "pushgp", "gp", "push"],
    author="Eddie Pantridge",
    author_email="erp12@hampshire.edu",
    license="LGPL",
    url="https://github.com/erp12/pyshgp",
    packages=find_packages(
        exclude=('examples', 'examples.*', 'tests', 'tests.*', 'docs', 'docs_source')
    ),
    classifiers=[
        "Development Status :: 4 - Beta",
        'Programming Language :: Python :: 3',
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    install_requires=[
        "numpy>=1.12.0",
        "scipy>=0.18.0",
        "scikit-learn>=0.18.0",
        "pandas>=0.23.4",
    ],
    setup_requires=[
        "pytest-runner",
        "flake8>=3.5.0",
        "flake8-docstrings>=1.3.0",
        "sphinx>=1.8.3",
        "m2r>=0.2.1",
        "mkdocs-nature>=0.3.1",
    ],
    tests_require=[
        "pytest"
    ],
)



"""The :mod:`instruction_set` module defines the InstructionSet class.

An InstructionSet is a collection of Instruction objects, stored by name. The
InstructionSet has methods to help define and register additional instructions.
"""
from typing import Sequence, Set
from itertools import chain
import re

from pyshgp.push.atoms import Instruction
from pyshgp.push.types import PushType
from pyshgp.push.instructions import common, numeric, text, code, io, logical


_CORE_INSTRUCTIONS = list(chain(
    common.instructions(),
    code.instructions(),
    numeric.instructions(),
    text.instructions(),
    io.instructions(),
    logical.instructions(),
))


class InstructionSet(dict):
    """A collection of Instruction objects stored by name.

    Parameters
    ----------
    register_all : bool, optional
        If True, all instructions in the core instruction set will be registered
        upon instanciation. Default is False.
    strip_docstrings : bool, optional
        If True, the docstring attribute of registered instructions will be
        removed to reduce memory footprint. Default is True.

    Attributes
    ----------
    register_all : bool, optional
        If True, all instructions in the core instruction set will be registered
        upon instanciation. Default is False.
    strip_docstrings : bool, optional
        If True, the docstring attribute of registered instructions will be
        removed to reduce memory footprint. Default is True.

    """

    def __init__(self, *, register_all: bool = False, strip_docstrings: bool = True):
        self.strip_docstrings = strip_docstrings
        if register_all:
            self.register_all()

    def register(self, instr: Instruction):
        """Register an Instruction object.

        Parameters
        ----------
        instr
            Instruction to register.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        if self.strip_docstrings and hasattr(instr, "docstring"):
            del instr.docstring
        self[instr.name] = instr
        return self

    def register_list(self, instrs: Sequence[Instruction]):
        """Register a list of Instruction ojbects.

        Parameters
        ----------
        instrs
            List of Instruction objects to register.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        for i in instrs:
            self.register(i)
        return self

    def register_by_type(self, type_names: Sequence[str], *, exclude: Sequence[str] = None):
        """Register all instructions that make use of the given type name.

        Parameters
        ----------
        type_names
            List of PushType names.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        for i in _CORE_INSTRUCTIONS:
            for type_name in type_names:
                i_types = i.relevant_types()
                if type_name in i_types:
                    include = True
                    if exclude is not None:
                        for i_type in i_types:
                            if i_type in exclude:
                                include = False
                    if include:
                        self.register(i)
                    break
        return self

    def register_by_name(self, name_pattern: str):
        """Register all instructions whose name match the given pattern.

        Parameters
        ----------
        name_pattern
            A regex string.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        re_pat = re.compile(name_pattern)
        for i in _CORE_INSTRUCTIONS:
            if re.match(re_pat, i.name) is not None:
                self.register(i)
        return self

    def register_all(self):
        """Register all core instructions defined in pyshgp.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        self.register_list(_CORE_INSTRUCTIONS)
        return self

    def register_n_inputs(self, n: int):
        """Create and register `n` input instructions.

        Parameters
        ----------
        n
            The number of input instructions to make.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        input_instructions = io.make_input_instructions(n)
        self.register_list(input_instructions)
        return self

    def unregister(self, instruction_name: str):
        """Unregister an instruction by name.

        Parameters
        ----------
        instruction_name
            The name of the instruciton to unregister.

        Returns
        -------
        InstructionSet
            A reference to the InstructionSet.

        """
        self.pop(instruction_name, None)
        return self

    def supported_types(self) -> Set[PushType]:
        """Return all PushTypes used through the registered instructions.

        Returns
        -------
        Set[PushType]
            The set of PushTypes that are used by the registered instructions.

        """
        all_types = set()
        for instr in self.values():
            all_types.update(instr.relevant_types())
        return all_types



"""The :mod:`atoms` module defines all types of Push Atoms.

An Atom is a peice of code in the Push language. A Literal Atom is a constant
value of one of the supported PushType. An Instruction Atom is a wrapped
function that can be used to manipulate a PushState. A CodeBlock is a sequence
of other Atoms is used to express nested expressions of code.
"""
from abc import ABC, abstractmethod
from typing import Any, Sequence
from itertools import chain, count
import json
from copy import copy

from pyshgp.push.types import PushType, push_type_by_name, push_type_of
from pyshgp.utils import JSONable, jsonify_collection


class Atom(JSONable):
    """Base class of all Atoms. The fundamental element of Push programs."""

    @classmethod
    def is_instance(cls, thing: Any) -> bool:
        """Return True if thing is Atom or subclass. Otherwise return False.

        Needed for when used as JIT PushType.

        Parameters
        ----------
        thing
            Anything.

        Returns
        -------
        bools
            Returns True if thing is an Atom or subclass of Atom. Returns False otherwise.

        """
        return isinstance(thing, cls)

    @classmethod
    def coerce(cls, thing: Any):
        """Convert thing into Atom by wrapping in Literal."""
        return Literal(thing)


class Closer(Atom):
    """An Atom dedicated to denoting the close of a CodeBlock in its flat representsion."""

    def jsonify(self) -> str:
        """Return the object as a JSON string."""
        return json.dumps({"a": "close"}, separators=(',', ':'))

    def __eq__(self, other):
        return isinstance(other, Closer)

    def __repr__(self) -> str:
        return "CLOSER"


class Literal(Atom):
    """An Atom which holds a constant value.

    Parameters
    ----------
    value : Any
        The value to be stored in Literal.
    push_type : PushType, optional
        The PushType of the Literal. Usually, the PushType's underlying native
        type is the same as the type of the Literal's value. The PushType is
        used to determine how to route the Literal onto the correct PushStack
        of the PushState during program execution. If push_type is none, the
        PushType will attempt to be infered from the type of the value.

    """

    __slots__ = ["push_type", "value"]

    def __init__(self, value: Any, push_type: PushType = None):
        self.push_type = push_type_of(value) if push_type is None else push_type
        self.value = self.push_type.coerce(value)

    def jsonify(self) -> str:
        """Return the object as a JSON string."""
        return json.dumps(
            {"a": "lit", "t": self.push_type.name, "v": self.value},
            separators=(',', ':')
        )

    def __eq__(self, other):
        if isinstance(other, Literal) and self.push_type == other.push_type:
            return self.value == other.value
        return False

    def __repr__(self) -> str:
        return str(self.value)


class Instruction(Atom, ABC):
    """An Atom which holds a function used to modify the PushStateself.

    The Instruction class is the abstract base class for specific implemenations
    that are configured differently. For example, see SimpleInstruction verses
    TakesStateInstruction.

    Parameters
    ----------
    name : str,
        A unique name for the instruction.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    Attributes
    ----------
    name : str,
        A unique name for the instruction.
    code_blocks : int
        The number of CodeBlocks to open following the instruction in a Genome.
    docstring : str, optional
        A string describing in the behavior of the Instruction.

    """

    __slots__ = ["name", "code_block", "docstring"]

    def __init__(self, name: str, code_blocks: int, docstring="Write me!"):
        self.name = name
        self.code_blocks = code_blocks
        self.docstring = docstring

    @abstractmethod
    def evaluate(self, push_state, interpreter_config=None):
        """Evaluate the instruction on the given PushState.

        Parameters
        ----------
        push_state: pyshgp.push.state.PushState
            The PushState to run the instrution on.
        interpreter_config: pyshgp.push.interpreter.PushInterpreterConfig
            The configuration of the Interpreter.

        Returns
        -------
        pyshgp.push.state.PushState
            Return the given state, possibly modified by the Instruction.

        """
        # Can't annotate types properly due to circular import.
        pass

    @abstractmethod
    def relevant_types(self) -> Sequence[str]:
        """Return a list of PushType names relevant to the instruction."""
        pass

    def jsonify(self) -> str:
        """Return the object as a JSON string."""
        return json.dumps({"a": "instr", "n": self.name}, separators=(',', ':'))

    def __eq__(self, other):
        if type(self) == type(other):
            return self.name == other.name
        return False

    def __hash__(self):
        return self.name.__hash__()

    def __repr__(self):
        return "Instruction<{n}>".format(n=self.name)


class JitInstructionRef(Atom):
    """Just-In-Time Instruction Reference.

    A JitInstructionRef is a placeholder atom. When a PushInterpreter evaluates
    a Just-In-Time Instruction Reference, it searches for an instruction with a
    specific name in its InstructionSet. If found, this instruction is
    evaluated by the PushInterpreter.

    A JitInstructionRef should be used when Instruction's function definition
    must refer to another instruction, including its own instruction.

    Parameters
    ----------
    name: str
        The name of the Instruction being referenced.

    """

    __slots__ = ["name"]

    def __init__(self, name: str):
        self.name = name

    def jsonify(self) -> str:
        """Return the object as a JSON string."""
        return json.dumps({"a": "jit-instr", "n": self.name}, separators=(',', ':'))

    def __eq__(self, other):
        if isinstance(other, JitInstructionRef):
            return self.name == other.name
        return False

    def __repr__(self):
        return "JitInstructionRef<{n}>".format(n=self.name)


class CodeBlock(list, Atom):
    """An Atom which holds a sequence of other Atoms."""

    def __init__(self, *args):
        for el in args:
            self._add(el)

    def _add(self, el):
        if isinstance(el, Atom):
            self.append(el)
        elif isinstance(el, list):
            self.append(CodeBlock.from_list(el))
        else:
            raise ValueError("Only Atoms can be added to CodeBlocks. Got {el}".format(el=el))

    @staticmethod
    def from_list(l: Sequence[Atom]):
        """Return a CodeBlock from a list of Atoms."""
        cb = CodeBlock()
        for el in l:
            cb._add(el)
        return cb

    @staticmethod
    def from_json_str(json_str: str, instruction_set: dict):
        """Create a CodeBlock from a JSON string."""
        # Can't annotate instruction_set type properly due to circular import.
        return AtomFactory.json_list_to_code_block(json.loads(json_str), instruction_set)

    def jsonify(self) -> str:
        """Return the object as a JSON string."""
        return jsonify_collection(self)

    def size(self, depth: int = 1) -> int:
        """Return the size of the block and the size of all the nested blocks."""
        if depth == 89:
            # Avoid RecursionError.
            return len(self)
        return sum([el.size(depth + 1) + 1 if isinstance(el, CodeBlock) else 1 for el in self])

    def depth(self) -> int:
        """Return the farthest depth of the CodeBlock."""
        i = iter(self)
        try:
            for level in count():
                i = chain([next(i)], i)
                i = chain.from_iterable(s for s in i if isinstance(s, Sequence) and not isinstance(s, str))
        except StopIteration:
            return level
        return 0

    def code_at_point(self, ndx: int) -> Atom:
        """Return a nested element of the CodeBlock using depth first traversal."""
        if ndx == 0:
            return self
        i = ndx
        for el in self:
            i = i - 1
            if i == 0:
                return el
            if isinstance(el, CodeBlock):
                next_depth = el.code_at_point(i)
                if next_depth is not None:
                    return next_depth
                i = i - el.size()

    def insert_code_at_point(self, code: Atom, index: int):
        """Insert an element into the CodeBlock using depth first traversal."""
        i = index
        for ndx, el in enumerate(self):
            if i == 0:
                self.insert(ndx, code)
                return self
            i = i - 1
            if isinstance(el, CodeBlock):
                next_depth = el.insert_code_at_point(code, i)
                if next_depth is not None:
                    return self
                i = i - el.size()

    def copy(self):
        """Copy the CodeBlock."""
        return copy(self)


class AtomFactory:
    """Produces specific types of Atoms from various sources.

    Any kinds of Atoms except CodeBlocks can be described in a dict. CodeBlocks
    can be described as lists of other Atom specifications. These representations
    are used for serialization and deserialization (JSON). The AtomFactory builds
    Atoms of any kind from these specifications.
    """

    # Can't annotate instruction_set type properly due to circular import.
    @staticmethod
    def json_dict_to_atom(json_dict: dict, instruction_set: dict) -> Atom:
        """Return the atom specified by the dict produced by JSON decoding."""
        atom_type = json_dict["a"]
        if atom_type == "close":
            return Closer()
        elif atom_type == "lit":
            push_type = push_type_by_name(json_dict["t"])
            value = push_type.coerce(json_dict["v"])
            return Literal(value, push_type)
        elif atom_type == "instr":
            return instruction_set[json_dict["n"]]
        elif atom_type == "jit-instr":
            return JitInstructionRef(json_dict["n"])
        else:
            raise ValueError("bad atom spec {s}".format(s=json_dict))

    # Can't annotate instruction_set type properly due to circular import.
    @staticmethod
    def json_list_to_code_block(json_list: list, instruction_set: dict) -> CodeBlock:
        """Return a CobdeBlock build from the list produced by JSON decoding."""
        cb = CodeBlock()
        for nested_atom_spec in json_list:
            if isinstance(nested_atom_spec, list):
                cb.append(AtomFactory.json_list_to_code_block(nested_atom_spec, instruction_set))
            else:
                cb.append(AtomFactory.json_dict_to_atom(nested_atom_spec, instruction_set))
        return cb

    # Can't annotate instruction_set type properly due to circular import.
    @staticmethod
    def json_str_to_atom_list(json_str: str, instruction_set: dict) -> Sequence[Atom]:
        """Return a list of Atoms parsed from the json_str."""
        atoms = []
        for atom_spec in json.loads(json_str):
            if isinstance(atom_spec, list):
                atoms.append(AtomFactory.json_list_to_code_block(atom_spec, instruction_set))
            else:
                atoms.append(AtomFactory.json_dict_to_atom(atom_spec, instruction_set))
        return atoms



"""The :mod:`interpreter` module defines the ``PushInterpreter`` class.

A ``PushInterpreter`` is capable of running Push programs and returning their
output. Push interpreters can be configured with a ``PushInterpreterConfig``
class to determine limits.
"""

from typing import Sequence, Union
import time
from enum import Enum

from pyshgp.push.state import PushState
from pyshgp.push.instruction_set import InstructionSet
from pyshgp.push.atoms import (
    Atom, Closer, Literal, Instruction, JitInstructionRef, CodeBlock
)
from pyshgp.utils import PushError


class PushInterpreterConfig:
    """Define a configuration for PushInterpreter.

    Parameters
    ----------
    atom_limit : int, optional
        Max number of atoms to process before terminating the execution of a
        given program. Default is 500
    growth_cap : int, optional
        Max number of elements that can be added to a PushState at any given
        step of program execution. If exceeded, program terminates. Default is
        500.
    runtime_limit : int, optional
        Max number of milliseconds to run a push program before forcing
        termination. Default is 2e10.
    reset_on_run : bool, optional
        If True, interpreter's state and status are reset every time a program
        is run. Default is True.


    Attributes
    ----------
    atom_limit : int
        Max number of atoms to process before terminating the execution of a
        given program. Default is 500
    growth_cap : int
        Max number of elements that can be added to a PushState at any given
        step of program execution. If exceeded, program terminates. Default is
        500.
    runtime_limit : int
        Max number of milliseconds to run a push program before forcing
        termination. Default is 2e10.
    reset_on_run : bool
        If True, interpreter's state and status are reset every time a program
        is run. Default is True.

    """

    # @TODO: Should PushInterpreterConfig be JSON serializable?
    # @TODO: Refactor atom_limit to step_limit

    def __init__(self, *, atom_limit=500, growth_cap=500, runtime_limit=2e10,
                 reset_on_run=True):
        self.atom_limit = atom_limit
        self.growth_cap = growth_cap
        self.runtime_limit = runtime_limit
        self.reset_on_run = reset_on_run


class PushInterpreterStatus(Enum):
    """Enum class of all potential statuses of a PushInterpreter."""

    normal = 1
    atom_limit_exceeded = 2
    runtime_limit_exceeded = 3
    growth_cap_exceeded = 4


class PushInterpreter:
    """An interpreter capable of running Push programs.

    Parameters
    ----------
    instruction_set : Union[InstructionSet, str], optional
        The InstructionSet to use for executing programs. Default is "core"
        which instansiates an InstructionSet using all the core instructions.
    config : PushInterpreterConfig, optional
        A PushInterpreterConfig specifying limits and early termination
        conditions. Default is None, which creates a config will all default
        values.

    Attributes
    ----------
    instruction_set : InstructionSet
        The InstructionSet to use for executing programs.
    config : PushInterpreterConfig
        A PushInterpreterConfig specifying limits and early termination
        conditions.
    state : PushState
        The current PushState. Contains one stack for each PushType utilized
        mentioned by the instructions in the instruction set.
    status : PushInterpreterStatus
        A string denoting if the Interpreter has enountered a situation
        where non-standard termination was required.

    """

    def __init__(self, instruction_set: Union[InstructionSet, str] = "core", config: PushInterpreterConfig = None):
        # If no instruction set given, create one and register all instructions.
        if instruction_set == "core":
            self.instruction_set = InstructionSet(register_all=True)
        else:
            self.instruction_set = instruction_set
        self._supported_types = self.instruction_set.supported_types()

        if config is None:
            self.config = PushInterpreterConfig()
        else:
            self.config = config

        # Initialize the PushState and status
        self.reset()

    def reset(self):
        """Reset the interpreter status and PushState."""
        self.state: PushState = PushState(self._supported_types)
        self.status: PushInterpreterStatus = PushInterpreterStatus.normal

    def _evaluate_instruction(self, instruction: Union[Instruction, JitInstructionRef]):
        self.state = instruction.evaluate(self.state, self.config)

    def evaluate_atom(self, atom: Atom):
        """Evaluate an Atom.

        Parameters
        ----------
        atom : Atom
            The Atom (Literal, Instruction, JitInstructionRef, or CodeBlock) to
            evaluate against the current PushState.

        """
        try:
            if isinstance(atom, Instruction):
                self._evaluate_instruction(atom)
            elif isinstance(atom, JitInstructionRef):
                self._evaluate_instruction(self.instruction_set[atom.name])
            elif isinstance(atom, CodeBlock):
                for a in atom[::-1]:
                    self.state["exec"].push(a)
            elif isinstance(atom, Literal):
                self.state[atom.push_type.name].push(atom.value)
            elif isinstance(atom, Closer):
                raise PushError("Closers should not be in push programs. Only genomes.")
            else:
                raise PushError("Cannont evaluate {t}, require a subclass of Atom".format(t=type(atom)))
        except (TypeError, ValueError) as e:
            err_type = type(e).__name__
            err_msg = str(e)
            raise PushError(
                "{t} raised while evaluating {atom}. Origional mesage: \"{m}\"".format(
                    t=err_type,
                    atom=atom,
                    m=err_msg
                )
            )

    def run(self, program: CodeBlock, inputs: Sequence,
            output_types: Sequence[str],
            verbose: bool = False):
        """Run a Push program given some inputs and desired output PushTypes.

        The general flow of this method is:
            1. Create a new push state
            2. Load the program and inputs.
            3. If the exec stack is empty, return the outputs.
            4. Else, pop the exec stack and process the atom.
            5. Return to step 3.

        Parameters
        ----------
        program
            Program to run.
        inputs
            A sequence of values to use as inputs to the push program.
        output_types
            A secence of values that denote the Pushtypes of the expected
            outputs of the push program.
        verbose
            If true, program execution steps will be printed. Default False.

        Returns
        -------
        Sequence
            A sequence of values pulled from the final push state. May contain
            pyshgp.utils.Token.no_stack_item if needed stacks are empty.

        """
        if self.config.reset_on_run:
            self.reset()

        self.state.load_program(program)
        self.state.load_inputs(inputs)
        stop_time = time.time() + self.config.runtime_limit
        steps = 0

        if verbose:
            print("Initial State:")
            self.state.pretty_print()
            print()

        while len(self.state["exec"]) > 0:
            if steps > self.config.atom_limit:
                self.status = PushInterpreterStatus.atom_limit_exceeded
                break
            if time.time() > stop_time:
                self.status = PushInterpreterStatus.runtime_limit_exceeded
                break

            next_atom = self.state["exec"].pop()

            if verbose:
                print("Current Atom:", next_atom)

            old_size = len(self.state)
            self.evaluate_atom(next_atom)
            if len(self.state) > old_size + self.config.growth_cap:
                self.status = PushInterpreterStatus.growth_cap_exceeded
                break

            if verbose:
                print("Current State:")
                self.state.pretty_print()
                print()
            steps += 1

        return self.state.observe_stacks(output_types)


DEFAULT_INTERPRETER = PushInterpreter()
